from tqdm import tqdm
import numpy as np

from dcmri import utils, pk_lib, lib, sig, pk, liver, tissue, kidney


def fake_aif(
        tacq=180.0,
        dt=1.5,
        BAT=20,
        field_strength=3.0,
        agent='gadodiamide',
        H=0.45,
        R10a=1/lib.T1(3.0, 'blood'),
        S0=150,
        model='SS',
        TR=0.005,
        FA=15,
        B1corr=1,
        TC=0.2,
        CNR=np.inf,
        dt_sim=0.1,
    ):
    """Synthetic AIF signal with noise.

    Args:
        tacq (float, optional): Duration of the acquisition in sec. Defaults to 180.
        dt (float, optional): Sampling interval in sec. Defaults to 1.5.
        BAT (int, optional): Bolus arrival time in sec. Defaults to 20.
        field_strength (float, optional): B0 field in T. Defaults to 3.0.
        agent (str, optional): Contrast agent generic name. Defaults to 'gadodiamide'.
        H (float, optional): Hematocrit. Defaults to 0.45.
        R10a (_type_, optional): Precontrast relaxation rate for blood in 1/sec. Defaults to 1/dc.T1(3.0, 'blood').
        S0 (int, optional): Signal scaling factor in blood (arbitrary units). Defaults to 150.
        model (str, optional): Scanning sequences, either steady-state ('SS') or saturation-recovery ('SR')
        TR (float, optional): Repetition time in sec. Defaults to 0.005.
        FA (int, optional): Flip angle. Defaults to 20.
        B1corr (float, optional): B1-correction factor. Defaults to 1.
        TC (float, optional): time to center of k-space in SR sequence. This is ignored when sequence='SS'. efaults to 0.2 sec.
        CNR (float, optional): Contrast-to-noise ratio, define as the ratio of signal-enhancement in the AIF to noise. Defaults to np.inf.
        dt_sim (float, optional): Sampling inteval of the forward modelling in sec. Defaults to 0.1.

    Returns: 
        tuple: time, aif, gt

        - **time**: array of time points.
        - **aif**: array of AIF signals.
        - **gt**: dictionary with ground truth values.
    """
    t = np.arange(0, tacq+dt, dt_sim)
    cp = pk_lib.aif_parker(t, BAT)
    rp = lib.relaxivity(field_strength, 'plasma', agent)
    R1b = R10a + rp*cp*(1-H)
    if model == 'SS':
        aif = sig.signal_ss(S0, R1b, TR, B1corr*FA)
    elif model == 'SR':
        aif = sig.signal_free(S0, R1b, TC, B1corr*FA)
    time = np.arange(0, tacq, dt)
    aif = utils.sample(time, t, aif, dt)
    sdev = (np.amax(aif)-aif[0])/CNR
    aif = utils.add_noise(aif, sdev)
    gt = {'t': t, 'cp': cp, 'cb': cp*(1-H),
          'TR': TR, 'FA': FA, 'B1corr':B1corr, 'S0b': S0}
    return time, aif, gt


# TODO align API with Tissue
def fake_brain(
        n=192,
        tacq=180.0,
        dt=1.5,
        BAT=20,
        Tav=8,
        field_strength=3.0,
        agent='gadodiamide',
        H=0.45,
        R10a=1/lib.T1(3.0, 'blood'),
        S0=150,
        model='SS',
        TR=0.005,
        FA=15,
        TC=0.2,
        CNR=np.inf,
        dt_sim=0.1,
        verbose=0,
    ):
    """Synthetic brain images data generated using Parker's AIF, the Shepp-Logan phantom and a two-compartment exchange tissue.

    Args:
        n (int, optional): matrix size. Defaults to 192.
        tacq (float, optional): Duration of the acquisition in sec. Defaults to 180.
        dt (float, optional): Sampling inteval in sec. Defaults to 1.5.
        BAT (int, optional): Bolus arrival time in sec. Defaults to 20.
        Tav (float, optional): Arterio-venous transit time. Defaults to 8 sec.
        field_strength (float, optional): B0 field in T. Defaults to 3.0.
        agent (str, optional): Contrast agent generic name. Defaults to 'gadodiamide'.
        H (float, optional): Hematocrit. Defaults to 0.45.
        R10a (_type_, optional): Precontrast relaxation rate for blood in 1/sec. Defaults to 1/dc.T1(3.0, 'blood').
        S0 (int, optional): Signal scaling factor for tissue (arbitrary units). Defaults to 150.
        model (str, optional): Scanning sequences, either steady-state ('SS') or saturation-recovery ('SR')
        TR (float, optional): Repetition time in sec. Defaults to 0.005.
        FA (int, optional): Flip angle. Defaults to 20.
        TC (float, optional): time to center of k-space in SR sequence. This is ignored when sequence='SS'. efaults to 0.2 sec.
        CNR (float, optional): Contrast-to-noise ratio, define as the ratio of signal-enhancement in the AIF to noise. Defaults to np.inf.
        dt_sim (float, optional): Sampling inteval of the forward modelling in sec. Defaults to 0.1.
        verbose (int, optional): if set to 0, no feedback is given during the computation. With verbose=1, an progress bar is shown. Defaults to 0.

    Returns: 
        tuple: time, signal, gt

        - **time**: array of time points.
        - **signal**: 3D array of pixel sigals, with dimensions (x,y,t).
        - **gt**: dictionary with ground truth values for concentrations and tissue parameters.

    Example:

        >>> import matplotlib.pyplot as plt
        >>> from matplotlib.animation import ArtistAnimation
        >>> import dcmri as dc

        Generate data:

        >>> time, signal, aif, gt = dc.fake_brain(n=64, CNR=100)

        Display as animation:

        >>> fig, ax = plt.subplots()
        >>> ims = []
        >>> for i in range(time.size):
        >>>     im = ax.imshow(signal[:,:,i], cmap='magma', animated=True, vmin=0, vmax=30)
        >>>     ims.append([im])
        >>> ani = ArtistAnimation(
        >>>     fig, ims, interval=50, blit=True,
        >>>     repeat_delay=0)
        >>> plt.show()

        .. image:: ../_static/animations/fake_brain.gif
            :alt: Example animation

    """
    # Parameter maps
    roi = lib.shepp_logan(n=n)
    im = lib.shepp_logan('T1', 'PD', 'Fb', 'vb', 'PS', 'vi', n=n)

    # Input
    t = np.arange(0, tacq+dt, dt_sim)
    cp = pk_lib.aif_parker(t, BAT)

    # Arterial signal
    rp = lib.relaxivity(field_strength, 'plasma', agent)
    R1b = R10a + rp*cp*(1-H)
    if model == 'SS':
        aif = sig.signal_ss(S0, R1b, TR, FA)
    elif model == 'SR':
        aif = sig.signal_free(S0, R1b, TC, FA)
    sdev = (np.amax(aif)-aif[0])/CNR
    time = np.arange(0, tacq, dt)
    aif = utils.sample(time, t, aif, dt)

    # Output
    conc = np.zeros((n, n, t.size))
    signal = np.zeros((n, n, time.size))
    signal_noisefree = np.zeros((n, n, time.size))

    # Tissue signal
    if verbose == 0:
        iter = range(n)
    else:
        iter = tqdm(range(n), desc='Generating fake brain data')

    for i in iter:
        for j in range(n):

            if roi['background'][i, j]:
                continue

            # Pixel concentration
            if roi['anterior artery'][i, j]:
                C = cp.copy()
            elif roi['sagittal sinus'][i, j]:
                C = pk.flux_comp(cp, Tav, dt=dt_sim)
            else:
                Fb = im['Fb'][i, j]
                vb = im['vb'][i, j]
                vi = im['vi'][i, j]
                PS = im['PS'][i, j]
                C = tissue.conc_tissue(
                    cp*(1-H), dt=dt_sim, kinetics='2CX', 
                    H=H, Fb=Fb, vb=vb, vi=vi, PS=PS)

            # Pixel signal
            R1 = 1/im['T1'][i, j] + rp*C
            if model == 'SS':
                s = sig.signal_ss(S0*im['PD'][i, j], R1, TR, FA)
            elif model == 'SR':
                s = sig.signal_spgr(S0*im['PD'][i, j], R1, TC, TR, FA)
            sig_noisefree = utils.sample(time, t, s, dt)
            s = utils.add_noise(sig_noisefree, sdev)

            # Save results
            conc[i, j, :] = C
            signal[i, j, :] = s
            signal_noisefree[i, j, :] = sig_noisefree

    gt = {'t': t, 'cp': cp, 'C': conc, 'cb': cp*(1-H),
          'signal': signal_noisefree,
          'Fp': im['Fb']*(1-H), 'vp': im['vb']*(1-H),
          'Fb': im['Fb'], 'vb': im['vb'],
          'PS': im['PS'], 'vi': im['vi'],
          'T1': im['T1'], 'PD': im['PD'],
          'TR': TR, 'FA': FA, 
          'S0': S0*im['PD']}

    gt['ve'] = gt['vi'] + gt['vp']

    # TODO: return the extracullar part of the distribution space
    # gt['ve'] = np.where(gt['PS'] > 0, gt['vp'] + gt['vi'], gt['vp'])

    return time, signal, aif, gt


def fake_tissue(
    tacq=180.0,
    dt=1.5,
    BAT=20,
    Fb=0.02,  
    vb=0.1,
    PS=0.003,  
    vi=0.2, 
    field_strength=3.0,
    agent='gadodiamide',
    H=0.45,
    R10a=1/lib.T1(3.0, 'blood'),
    R10=1/lib.T1(3.0, 'muscle'),
    S0b=100,
    S0=150,
    model='SS',
    TR=0.005,
    FA=15,
    TC=0.2,
    CNR=np.inf,
    dt_sim=0.1,
):
    """Synthetic data generated using Parker's AIF and a two-compartment exchange tissue.

    Args:
        tacq (float, optional): Duration of the acquisition in sec. Defaults to 180.
        dt (float, optional): Sampling interval in sec. Defaults to 1.5.
        BAT (int, optional): Bolus arrival time in sec. Defaults to 20.
        Fb (float, optional): Blood flow in mL/sec/mL. Defaults to 0.1.
        vb (float, optional): Blood volume fraction. Defaults to 0.05.
        PS (float, optional): Permeability-surface area product in 1/sec. Defaults to 0.003.
        vi (float, optional): Interstitial volume fraction. Defaults to 0.3.
        field_strength (float, optional): B0 field in T. Defaults to 3.0.
        agent (str, optional): Contrast agent generic name. Defaults to 'gadodiamide'.
        H (float, optional): Hematocrit. Defaults to 0.45.
        R10a (_type_, optional): Precontrast relaxation rate for blood in 1/sec. Defaults to 1/dc.T1(3.0, 'blood').
        R10 (int, optional): Precontrast relaxation rate for tissue in 1/sec. Defaults to 1.
        S0b (int, optional): Signal scaling factor for blood (arbitrary units). Defaults to 100.
        S0 (int, optional): Signal scaling factor for tissue (arbitrary units). Defaults to 150.
        model (str, optional): Scanning sequences, either steady-state ('SS') or saturation-recovery ('SR')
        TR (float, optional): Repetition time in sec. Defaults to 0.005.
        FA (int, optional): Flip angle. Defaults to 20.
        TC (float, optional): time to center of k-space in SR sequence. This is ignored when sequence='SS'. efaults to 0.2 sec.
        CNR (float, optional): Contrast-to-noise ratio, define as the ratio of signal-enhancement in the AIF to noise. Defaults to np.inf.
        dt_sim (float, optional): Sampling inteval of the forward modelling in sec. Defaults to 0.1.

    Returns: 
        tuple: time, aif, roi, gt

        - **time**: array of time points.
        - **aif**: array of AIF signals.
        - **roi**: array of ROI signals.
        - **gt**: dictionary with ground truth values for concentrations and tissue parameters.
    """
    t = np.arange(0, tacq+dt, dt_sim)
    cp = pk_lib.aif_parker(t, BAT)
    C = tissue.conc_tissue(cp*(1-H), dt=dt_sim, kinetics='2CX',
                       H=H, Fb=Fb, vb=vb, PS=PS, vi=vi)
    rp = lib.relaxivity(field_strength, 'plasma', agent)
    R1b = R10a + rp*cp*(1-H)
    R1 = R10 + rp*C
    if model == 'SS':
        aif = sig.signal_ss(S0b, R1b, TR, FA)
        roi = sig.signal_ss(S0, R1, TR, FA)
    elif model == 'SR':
        aif = sig.signal_free(S0b, R1b, TC, FA)
        roi = sig.signal_spgr(S0, R1, TC, TR, FA)
    time = np.arange(0, tacq, dt)
    aif = utils.sample(time, t, aif, dt)
    roi = utils.sample(time, t, roi, dt)
    sdev = (np.amax(aif)-aif[0])/CNR
    aif = utils.add_noise(aif, sdev)
    roi = utils.add_noise(roi, sdev)
    gt = {'t': t, 'cp': cp, 'C': C, 'cb': cp*(1-H),
          'Fb': Fb, 'vb': vb, 
          'Fp': Fb*(1-H), 'vp': vb*(1-H), 'PS': PS, 'vi': vi,
          'Ktrans': Fb*(1-H)*PS/(Fb*(1-H)+PS),
          'TR': TR, 'FA': FA, 'S0': S0}
    return time, aif, roi, gt


def fake_liver(
    tacq = 180.0,
    dt = 1.5,
    BAT = 20,
    Tg = 10,
    H = 0.45,
    ve = 0.3,
    Fp = 0.01,
    fa = 0.2,
    Ta = 2,
    khe = 0.003,
    Th = 20*60,  
    field_strength = 3.0,
    agent = 'gadoxetate',
    R10a = 1/lib.T1(3.0, 'blood'),
    R10 = 1/lib.T1(3.0, 'liver'),
    S0b = 100,
    S0 = 150,
    sequence = 'SS',
    TR = 0.005,
    FA = 15,
    TC = 0.5,
    CNR = np.inf,
    dt_sim = 0.1,
):
    """Synthetic data for liver tissue.

    Args:
        tacq (float, optional): Duration of the acquisition in sec. 
          Defaults to 180.
        dt (float, optional): Sampling interval in sec. Defaults to 1.5.
        BAT (float, optional): Bolus arrival time in sec. Defaults to 20.
        Tg (float, optional): Gut mean transit time. Defaults to 10 sec.
        H (float, optional): Hematocrit. Defaults to 0.45.
        ve (float, optional): Extracellular volume fraction. Defaults to 0.3.
        Fp (float, optional): Plasma flow in mL/sec/mL. Defaults to 0.01.
        fa (float, optional): Arterial flow fraction. Defaults to 0.2.
        Ta (float, optional): Arterial mean transit time. Defaults to 2 sec.
        khe (float, optional): Hepatocellular uptake rate. Defaults to 0.003.
        Th (float, optional): Hepatocyte mean transit time. Defaults to 20 
          mins.
        field_strength (float, optional): B0 field in T. Defaults to 3.0.
        agent (str, optional): Contrast agent generic name. Defaults to 
          'gadoxetate'.
        R10a (_type_, optional): Precontrast relaxation rate for blood in 
          1/sec. Defaults to 1/dc.T1(3.0, 'blood').
        R10 (int, optional): Precontrast relaxation rate for tissue in 1/sec. 
          Defaults to 1.
        S0b (int, optional): Signal scaling factor for blood (arbitrary 
          units). Defaults to 100.
        S0 (int, optional): Signal scaling factor for tissue (arbitrary 
          units). Defaults to 150.
        sequence (str, optional): Scanning sequences, either steady-state 
          ('SS') or Steady-state with inflow effects ('SSI').
        TR (float, optional): Repetition time in sec. Defaults to 0.005.
        FA (int, optional): Flip angle. Defaults to 20.
        TC (float, optional): Aortic inflow time for SSI sequence model. 
          This is ignored when sequence='SS'. Defaults to 0.2 sec.
        CNR (float, optional): Contrast-to-noise ratio, define as the ratio 
          of signal-enhancement in the AIF to noise. Defaults to np.inf.
        dt_sim (float, optional): Sampling inteval of the forward modelling 
          in sec. Defaults to 0.1.

    Returns: 
        tuple: time, aif, roi, gt

        - **time**: array of time points.
        - **aif**: array of AIF signals.
        - **vif**: array of VIF signals.
        - **roi**: array of ROI signals.
        - **gt**: dictionary with ground truth values for concentrations and 
          tissue parameters.
    """
    t = np.arange(0, tacq+dt, dt_sim)
    cp = pk_lib.aif_parker(t, BAT)
    cv = pk.flux_comp(cp, Tg, t)
    C = liver.conc_liver(cp*(1-H), dt=dt_sim, cv=cv*(1-H), sum=False,
                      H=H, ve=ve, Fp=Fp, fa=fa, Ta=Ta, khe=khe, Th=Th)
    rp = lib.relaxivity(field_strength, 'plasma', agent)
    rh = lib.relaxivity(field_strength, 'hepatocytes', agent)
    R1a = R10a + rp*cp*(1-H)
    R1v = R10a + rp*cv*(1-H)
    R1 = R10 + rp*C[0, :] + rh*C[1, :]
    if sequence == 'SS':
        aif = sig.signal_ss(S0b, R1a, TR, FA)
        vif = sig.signal_ss(S0b, R1v, TR, FA)
        roi = sig.signal_ss(S0, R1, TR, FA)
    elif sequence == 'SSI':
        aif = sig.signal_spgr(S0b, R1a, TC, TR, FA, n0=1)
        vif = sig.signal_ss(S0b, R1v, TR, FA)
        roi = sig.signal_ss(S0, R1, TR, FA)
    time = np.arange(0, tacq, dt)
    aif = utils.sample(time, t, aif, dt)
    vif = utils.sample(time, t, vif, dt)
    roi = utils.sample(time, t, roi, dt)
    sdev = (np.amax(aif)-aif[0])/CNR
    aif = utils.add_noise(aif, sdev)
    vif = utils.add_noise(vif, sdev)
    roi = utils.add_noise(roi, sdev)
    gt = {'t': t, 'cp': cp, 'cv':cv*(1-H), 
          'C': np.sum(C,axis=0), 'cb': cp*(1-H),
          've': ve, 'Fp': Fp, 
          'Fb': Fp/(1-H), 'fa': fa, 'Ta': Ta, 'khe': khe, 'Th': Th,
          'TR': TR, 'FA': FA, 'S0': S0}
    return time, aif, vif, roi, gt


def fake_tissue2scan(
    tacq=180.0,
    tbreak=60,
    dt=1.5,
    BAT=20,
    Fb=0.1,
    vb=0.05,
    PS=0.003,
    vi=0.2,
    field_strength=3.0,
    agent='gadodiamide',
    H=0.45,
    R10a=1/lib.T1(3.0, 'blood'),
    R10=1/lib.T1(3.0, 'muscle'),
    S0b1=100,
    S01=150,
    S0b2=200,
    S02=300,
    model='SS',
    TR=0.005,
    FA=15,
    TC=0.2,
    CNR=np.inf,
    dt_sim=0.1,
):
    """Synthetic data generated using Parker's AIF, a two-compartment exchange tissue and a steady-state sequence.

    Args:
        tacq (float, optional): Duration of the acquisition in sec. Defaults to 180.
        tbreak (float, optional): Break time between the two scans in sec. Defaults to 60.
        dt (float, optional): Sampling inteval in sec. Defaults to 1.5.
        BAT (int, optional): Bolus arrival time in sec. Defaults to 20.
        Fb (float, optional): Blood flow in mL/sec/mL. Defaults to 0.1.
        vb (float, optional): Blood volume fraction. Defaults to 0.05.
        PS (float, optional): Permeability-surface area product in 1/sec. Defaults to 0.003.
        vi (float, optional): Interstitial volume fraction. Defaults to 0.3.
        field_strength (float, optional): B0 field in T. Defaults to 3.0.
        agent (str, optional): Contrast agent generic name. Defaults to 'gadodiamide'.
        H (float, optional): Hematocrit. Defaults to 0.45.
        R10a (_type_, optional): Precontrast relaxation rate for blood in 1/sec. Defaults to 1/dc.T1(3.0, 'blood').
        R10 (int, optional): Precontrast relaxation rate for tissue in 1/sec. Defaults to 1.
        S0b1 (int, optional): Signal scaling factor for blood in the first scan (arbitrary units). Defaults to 100.
        S01 (int, optional): Signal scaling factor for tissue in the first scan (arbitrary units). Defaults to 150.
        S0b2 (int, optional): Signal scaling factor for blood in the second scan (arbitrary units). Defaults to 100.
        S02 (int, optional): Signal scaling factor for tissue in the second scan (arbitrary units). Defaults to 150.
        model (str, optional): Scanning sequences, either steady-state ('SS') or saturation-recovery ('SR')
        TR (float, optional): Repetition time in sec. Defaults to 0.005.
        FA (int, optional): Flip angle. Defaults to 15.
        TC (float, optional): time to center of k-space in SR sequence. This is ignored when sequence='SS'. Defaults to 0.2 sec.
        CNR (float, optional): Contrast-to-noise ratio, define as the ratio of signal-enhancement in the AIF to noise. Defaults to np.inf.
        dt_sim (float, optional): Sampling inteval of the forward modelling in sec. Defaults to 0.1.

    Returns: 
        tuple: time, aif, roi, gt

        - **time**: tuple with two arrays of time points.
        - **aif**: tuple with two arrays of AIF signals.
        - **roi**: tuple with two arrays of ROI signals.
        - **gt**: dictionary with ground truth values for concentrations and tissue parameters.
    """
    # Simulate relaxation rates over the full time range
    t = np.arange(0, 2*tacq+tbreak+dt, dt_sim)
    cp = pk_lib.aif_parker(t, BAT)
    cp += pk_lib.aif_parker(t, tacq+tbreak+BAT)
    C = tissue.conc_tissue(cp*(1-H), dt=dt_sim, kinetics='2CX',
                       H=H, Fb=Fb, vb=vb, PS=PS, vi=vi)
    rp = lib.relaxivity(field_strength, 'plasma', agent)
    R1b = R10a + rp*cp*(1-H)
    R1 = R10 + rp*C

    # Generate the signals from the first scan
    if model == 'SS':
        aif = sig.signal_ss(S0b1, R1b, TR, FA)
        roi = sig.signal_ss(S01, R1, TR, FA)
    elif model == 'SR':
        aif = sig.signal_free(S0b1, R1b, TC, FA)
        roi = sig.signal_spgr(S01, R1, TC, TR, FA)
    time1 = np.arange(0, tacq, dt)
    aif1 = utils.sample(time1, t, aif, dt)
    roi1 = utils.sample(time1, t, roi, dt)
    sdev = (np.amax(aif1)-aif1[0])/CNR
    aif1 = utils.add_noise(aif1, sdev)
    roi1 = utils.add_noise(roi1, sdev)

    # Generate the second signals
    if model == 'SS':
        aif = sig.signal_ss(S0b2, R1b, TR, FA)
        roi = sig.signal_ss(S02, R1, TR, FA)
    elif model == 'SR':
        aif = sig.signal_free(S0b2, R1b, TC, FA)
        roi = sig.signal_spgr(S02, R1, TC, TR, FA)
    time2 = np.arange(tacq+tbreak, 2*tacq+tbreak, dt)
    aif2 = utils.sample(time2, t, aif, dt)
    roi2 = utils.sample(time2, t, roi, dt)
    sdev = (np.amax(aif2)-aif2[0])/CNR
    aif2 = utils.add_noise(aif2, sdev)
    roi2 = utils.add_noise(roi2, sdev)

    # Build return values
    time = (time1, time2)
    aif = (aif1, aif2)
    roi = (roi1, roi2)
    gt = {'t': t, 'cp': cp, 'C': C, 'cb': cp*(1-H),
          'Fb': Fb, 'vb': vb, 'PS': PS, 'vi': vi,
          'Fp': Fb*(1-H), 'vp':vb*(1-H),
          'TR': TR, 'FA': FA, 'S01': S01, 'S02': S02}
    return time, aif, roi, gt


def fake_kidney(
    tacq=180.0,
    dt=1.5,
    BAT=20,
    Fp=0.03,
    Eg=0.15,
    fc=0.8,
    Tg=4,
    Tv=10,
    Tpt=60,
    Tlh=60,
    Tdt=30,
    Tcd=30,
    field_strength=3.0,
    agent='gadoterate',
    Hct=0.45,
    R10a=1/lib.T1(3.0, 'blood'),
    R10c=1/lib.T1(3.0, 'kidney'),
    R10m=1/lib.T1(3.0, 'kidney'),
    S0b=100,
    S0=150,
    model='SR',
    TC=0.2,
    TR=0.005,
    FA=15,
    CNR=np.inf,
    dt_sim=0.1,
):
    """Synthetic data generated using Parker's AIF, and a multicompartment kidney model.

    Args:
        tacq (float, optional): Duration of the acquisition in sec. Defaults to 180.
        dt (float, optional): Sampling inteval in sec. Defaults to 1.5.
        BAT (float, optional): Bolus arrival time in sec. Defaults to 20.
        Fp (float, optional): Plasma flow in mL/sec/mL. Defaults to 0.03.
        Eg (float, optional): Glomerular extraction fraction. Defaults to 0.15.
        fc (float, optional): Cortical flow fraction. Defaults to 0.8.
        Tg (float, optional): Glomerular mean transit time in sec. Defaults to 4.
        Tv (float, optional): Peritubular & venous mean transit time in sec. Defaults to 10.
        Tpt (float, optional): Proximal tubuli mean transit time in sec. Defaults to 60.
        Tlh (float, optional): Lis of Henle mean transit time in sec. Defaults to 60.
        Tdt (float, optional): Distal tubuli mean transit time in sec. Defaults to 30.
        Tcd (float, optional): Collecting duct mean transit time in sec. Defaults to 30.
        field_strength (float, optional): B0 field in T. Defaults to 3.0.
        agent (str, optional): Contrast agent generic name. Defaults to 'gadodiamide'.
        Hct (float, optional): Hematocrit. Defaults to 0.45.
        R10a (float, optional): Precontrast relaxation rate for blood in 1/sec. Defaults to 1/dc.T1(3.0, 'blood').
        R10c (float, optional): Precontrast relaxation rate for cortex in 1/sec. Defaults to 1.
        R10m (float, optional): Precontrast relaxation rate for cortex in 1/sec. Defaults to 1.
        S0b (float, optional): Signal scaling factor for blood (arbitrary units). Defaults to 100.
        S0 (float, optional): Signal scaling factor for tissue (arbitrary units). Defaults to 150.
        TC (float, optional): Time to readout of the k-space center in sec. Defaults to 0.2.
        TR (float, optional): Repetition time in sec. Defaults to 0.005.
        FA (float, optional): Flip angle. Defaults to 15.
        CNR (float, optional): Contrast-to-noise ratio, define as the ratio of signal-enhancement in the AIF to noise. Defaults to np.inf.
        dt_sim (float, optional): Sampling interval of the forward modelling in sec. Defaults to 0.1.

    Returns: 
        tuple: time, aif, roi, gt

        - **time**: array of time points.
        - **aif**: array of AIF signals.
        - **roi**: tuple with arrays of cortex and medulla signals.
        - **gt**: dictionary with ground truth values for concentrations.
    """
    t = np.arange(0, tacq+dt, dt_sim)
    cp = pk_lib.aif_parker(t, BAT)
    Cc, Cm = kidney.conc_kidney_cm(
        cp, Fp, Eg, fc, Tg, Tv, Tpt, Tlh, Tdt, Tcd, dt=dt_sim, kinetics='7C')
    rp = lib.relaxivity(field_strength, 'plasma', agent)
    R1b = R10a + rp*cp*(1-Hct)
    R1c = R10c + rp*Cc
    R1m = R10m + rp*Cm
    if model == 'SS':
        aif = sig.signal_ss(S0b, R1b, TR, FA)
        roic = sig.signal_ss(S0, R1c, TR, FA)
        roim = sig.signal_ss(S0, R1m, TR, FA)
    elif model == 'SR':
        aif = sig.signal_free(S0b, R1b, TC, FA)
        roic = sig.signal_spgr(S0, R1c, TC, TR, FA)
        roim = sig.signal_spgr(S0, R1m, TC, TR, FA)
    time = np.arange(0, tacq, dt)
    aif = utils.sample(time, t, aif, dt)
    roic = utils.sample(time, t, roic, dt)
    roim = utils.sample(time, t, roim, dt)
    sdev = (np.amax(aif)-aif[0])/CNR
    aif = utils.add_noise(aif, sdev)
    roic = utils.add_noise(roic, sdev)
    roim = utils.add_noise(roim, sdev)
    gt = {'t': t, 'cp': cp, 'Cc': Cc, 'Cm': Cm, 'cb': cp*(1-Hct)}
    return time, aif, (roic, roim), gt



import numpy as np
from scipy.integrate import trapezoid

import dcmri.pk as pk
import dcmri.lib as lib


def aif_tristan(
    t: np.ndarray,
    agent='gadoterate',
    dose=0.2,
    rate=3,
    BAT=0,
    weight=73,
    CO=97,
    E=0.07,
    Thl=13,
    Dhl=0.5,
    Tp=25,
    Te=350,
    Ee=0.18,
    dtol=0.01,
) -> np.ndarray:
    """Arterial input function with default parameters for young healthy 
    volunteers.

    This AIF was measured in the TRISTAN project (Min et al 2024). The default 
    values are for young healthy volunteers but since the AIF is built on a 
    whole-body model of the circulation, they can be modified to generate 
    virtual populations. 

    Reference:

    Thazin Min, Marta Tibiletti, Paul Hockings, Aleksandra Galetin, 
    Ebony Gunwhy, Gerry Kenna, Nicola Melillo, Geoff JM Parker, 
    Gunnar Schuetz, Daniel Scotcher, John Waterton, Ian Rowe, and 
    Steven Sourbron. *Measurement of liver function with dynamic 
    gadoxetate-enhanced MRI: a validation study in healthy volunteers*. 
    Proc Intl Soc Mag Reson Med, Singapore 2024.

    Args:
        t (np.ndarray): Array of time points
        agent (str, optional): Contrast agent generic name. Defaults to 
          'gadoterate'.
        dose (float, optional): Contrast agent dose in mL/kg. Defaults to 0.2.
        rate (float, optional): Contrast agent injection rate in mL/sec. 
          Defaults to 3.
        BAT (float, optional): Bolus arrival time in sec. Defaults to 0.
        weight (float, optional): Subject weight in kg. Defaults to 73.
        CO (float, optional): Cardiac output in mL/sec. Defaults to 97.
        E (float, optional): Body extraction fraction. Defaults to 0.07.
        Thl (float, optional): Mean transit time of the heart-lung system in 
          sec. Defaults to 13.
        Dhl (float, optional): Transit time dispersion of the heart-lung 
          system. Defaults to 0.5.
        Tp (float, optional): Plasma mean transit time in sec of the other 
          organs. Defaults to 25.
        Te (float, optional): Extravascular mean transit time in sec. 
          Defaults to 350.
        Ee (float, optional): Extraction fraction into the extravascular 
          space. Defaults to 0.18.
        dtol (float, optional): Dose tolerance. Defaults to 0.01.

    Returns:
        np.ndarray: Aorta blood concentrations in mmol/mL.

    Example:

        Generate AIFs with different levels of cardiac output:

    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import dcmri as dc

        # Time points in sec
        t = np.arange(0, 180, 2.0)

        # Plot aifs with different levels of cardiac output, including the 
        # default of 100mL/sec.
        plt.plot(t/60, 1000*dc.aif_tristan(t, BAT=30), 'r-', label='AIF (default)')
        plt.plot(t/60, 1000*dc.aif_tristan(t, BAT=30, CO=150), 'g-', 
                 label='AIF (increased cardiac output)')
        plt.plot(t/60, 1000*dc.aif_tristan(t, BAT=30, CO=75), 'b-', 
                 label='AIF (reduced cardiac output)')
        plt.xlabel('Time (min)')
        plt.ylabel('Concentration (mmol/mL)')
        plt.legend()
        plt.show()    
    """
    conc = lib.ca_conc(agent)
    Ji = lib.ca_injection(t, weight,conc, dose, rate, BAT)
    Jb = flux_aorta(Ji, t, E=E,
                    heartlung=['chain', (Thl, Dhl)],
                    organs=['2cxm', ([Tp, Te], Ee)],
                    tol=dtol)
    return Jb/CO


def flux_aorta(J_vena: np.ndarray,
               t=None, dt=1.0, E=0.1, FFkl=0.0, FFk=0.5,
               heartlung=['pfcomp', (10, 0.2)],
               organs=['2cxm', ([20, 120], 0.15)],
               kidneys=['comp', (10,)],
               liver=['pfcomp', (10, 0.2)],
               tol=0.001):
    
    """Whole-body model for indicator flux through the aorta.

    See section :ref:`whole-body-tissues` for a more detailed description of 
    this model.

    Args:
        J_vena (np.ndarray): Indicator influx (mmol/sec) into the veins. 
        t (np.ndarray, optional): Array of time points (sec), must be of 
          equal size as J_vena. If not provided, the time points are uniformly 
          sampled with interval dt. Defaults to None.
        dt (float, optional): Sampling interval in sec. Defaults to 1.0.
        E (float, optional): Body extraction fraction. Defaults to 0.1.
        FFkl (float, optional): Fraction of the cardiac output that passes 
          through kidney and liver. Set FFkl=0 for a whole body model without 
          explicit kidney and liver spaces. Defaults to 0.0.
        FFk (float, optional): Kidney fraction of the flow to kidney and liver. 
          With FFk=0, only the liver is modelled. With FFk=1, only the kidneys 
          are modelled. Defaults to 0.5.
        heartlung (list): 3-element list specifying the model to use for the 
          heart-lung system (see notes for detail).
        organs (list): 3-element list specifying the model to use for the 
          organs (see notes for detail). 
        kidneys (list): 3-element list specifying the model to use for the 
          kidneys (see notes for detail). This keyword is ignored if FFkl=0 
          or FFk=0.
        liver (list): 3-element list specifying the model to use for the 
          liver (see notes for detail). This keyword is ignored if FFkl=0 
          or FFk=1.
        tol (float, optional): Dose tolerance in the solution. The solution 
          propagates the input through the system, until the dose that is 
          left in the system is given by tol*dose0, where dose0 is the 
          initial dose. Defaults to 0.001.

    Returns:
        tuple: Indicator fluxes (mmol/sec) through the vena cava and aorta.

    Notes:

        The lists specifying each organ system consist of 3 elements: the 
        model (str), its parameters (tuple) and any keyword parameters (dict). 
        Any of the basic pharmacokinetic blocks can be used. For instance,
        *chain*, *plug-flow compartment*, and *compartment* would be specified 
        as follows:

          - **chain**: ['chain', (Thl, Dhl), {'solver':'step'}]
          - **plug-flow compartment**: ['pfcomp', (Thl, Dhl), {'solver':'interp'}}
          - **compartment**: ['comp', Thl]
          - **2cxm**: ['2cxm', ([To, Te], Eo)]


    Example:

        Generate flux through aorta:

    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import dcmri as dc

        # Generate a stepwise injection:
        t = np.arange(0, 120, 2.0)
        Ji = dc.ca_injection(t, 70, 0.5, 0.2, 3, 30)

        # Calculate the fluxes in mmol/sec:
        Ja = dc.flux_aorta(Ji, t)

        # Plot the fluxes:
        plt.plot(t/60, Ja, 'r-', label='Aorta')
        plt.xlabel('Time (min)')
        plt.ylabel('Indicator flux (mmol/sec)')
        plt.legend()
        plt.show()
    """

    dose = trapezoid(J_vena, x=t, dx=dt)
    min_dose = tol*dose

    # Residuals of each pathway
    Rk = FFk*FFkl*(1-E)
    Rl = (1-FFk)*FFkl*(1-E)
    Ro = (1-FFkl)*(1-E)

    # Initialize output
    J_aorta_total = np.zeros(J_vena.size)

    while dose > min_dose:

        # Aorta flux of the current pass
        J_aorta = pk.flux(
            J_vena, *heartlung[1], t=t, dt=dt, model=heartlung[0])

        # Add to the total aorta flux
        J_aorta_total += J_aorta

        # Venous flux of the current pass
        J_vena = Ro * pk.flux(
            J_aorta, *organs[1], t=t, dt=dt, model=organs[0])
        if Rl > 0:
            J_vena += Rl * pk.flux(
                J_aorta, *liver[1], t=t, dt=dt, model=liver[0])
        if Rk > 0:
            J_vena += Rk * pk.flux(
                J_aorta, *kidneys[1], t=t, dt=dt, model=kidneys[0])

        # Get residual dose in current pass
        dose = trapezoid(J_vena, x=t, dx=dt)

    return J_aorta_total



import numpy as np
import pydmr
import dcmri as dc


# Debugging mode
VERBOSE = 1
SHOW = True

VERBOSE = 0
SHOW = False


def test_ui_kidney():
    time, aif, roi, gt = dc.fake_tissue(R10=1/dc.T1(3.0,'kidney'))
    #
    # Override the parameter defaults to match the experimental conditions of 
    # the synthetic test data:
    #
    params = {
        'aif':aif,
        'dt':time[1],
        'agent': 'gadodiamide',
        'TR': 0.005,
        'FA': 15,
        'TS': time[1],
        'R10': 1/dc.T1(3.0,'kidney'),
        't0': 10*time[1],
    }
    #
    # Train a two-compartment filtration model on the ROI data and plot the fit:
    #
    model = dc.Kidney(kinetics='2CF', **params)
    model.train(time, roi)
    model.plot(time, roi, ref=gt, show=SHOW)
    assert model.cost(time, roi) < 0.5
    assert 80 < model.params('Tt', round_to=0) < 90


def test_ui_kidney_cort_med():

    time, aif, roi, gt = dc.fake_kidney()

    model = dc.KidneyCortMed(
        aif = aif,
        dt = time[1],
        agent = 'gadoterate',
        TR = 0.005,
        FA = 15,
        TC = 0.2,
        TS = time[1],
        n0 = 10,
    )

    model.train(time, roi)
    model.plot(time, roi, ref=gt, show=SHOW)

    assert model.cost(time, roi) < 1
    assert model.params('Tdt', round_to=0) == 27


def test_ui_aorta_kidneys():

    datafile = dc.fetch('minipig_renal_fibrosis')
    dmr = pydmr.read(datafile, 'nest')
    rois, pars = dmr['rois']['Pig']['Test'], dmr['pars']['Pig']['Test']

    aorta_kidneys = dc.AortaKidneys(

        # Configuration
        sequence='SSI',
        heartlung='chain',
        organs='comp',
        agent="gadoterate",

        # General parameters
        field_strength=pars['B0'],
        t0=pars['TS']*pars['n0'], 

        # Injection protocol
        weight=pars['weight'],
        dose=pars['dose'],
        rate=pars['rate'],

        # Sequence parameters
        TR=pars['TR'],
        FA=pars['FA'],
        TS=pars['TS'],

        # Aorta parameters
        CO=60,  
        R10a=1/dc.T1(pars['B0'], 'blood'),

        # Kidney parameters
        vol_lk=85,
        vol_rk=85,
        R10_lk=1/dc.T1(pars['B0'], 'kidney'),
        R10_rk=1/dc.T1(pars['B0'], 'kidney'),
    )

    # Define time and signal data
    time = pars['TS'] * np.arange(len(rois['Aorta']))
    t = (time, time, time)
    signal = (rois['Aorta'], rois['LeftKidney'], rois['RightKidney'])

    # Train model and show result
    aorta_kidneys.train(t, signal)
    aorta_kidneys.plot(t, signal, show=SHOW)
    aorta_kidneys.print_params(round_to=4)
    Fp_lk = aorta_kidneys.params('Fp_lk') # 0.032
    assert 0.030 < np.round(Fp_lk, 3) < 0.033



if __name__ == "__main__":

    # test_ui_kidney()
    # test_ui_kidney_cort_med()
    test_ui_aorta_kidneys()

    print('All ui_kidney tests passed!!')




from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np

import dcmri.ui as ui
import dcmri.lib as lib
import dcmri.sig as sig
import dcmri.utils as utils
import dcmri.pk_aorta as pk_aorta


class Aorta(ui.Model):
    """Whole-body model for the aorta signal.

    This model uses a whole body model to predict the signal in the 
    aorta (see :ref:`whole-body-tissues`). 

    See Also:
        `AortaKidneys`, `AortaLiver`

    Args:
        organs (str, optional): Model for the organs in the whole-body 
          model. The options are 'comp' (one compartment) and '2cxm' 
          (two-compartment exchange). Defaults to 'comp'.
        heartlung (str, optional): Model for the heart-lung system in 
          the whole-body model. Options are 'comp' (compartment), 
          'pfcomp' (plug-flow compartment) or 'chain'. Defaults to 
          'pfcomp'.
        sequence (str, optional): imaging sequence model. Possible 
          values are 'SS' (steady-state), 'SR' (saturation-recovery), 
          'SSI' (steady state with inflow correction) and 'lin' 
          (linear). Defaults to 'SS'.
        params (dict, optional): values for the model parameters,
          specified as keyword parameters. Defaults are used for any 
          that are not provided. See table 
          :ref:`Aorta-defaults` for a list of parameters and 
          their default values.

    Notes:

        In the table below, if **Bounds** is None, the parameter is fixed 
        during training. Otherwise it is allowed to vary between the 
        bounds given.

        .. _Aorta-defaults:
        .. list-table:: Aorta parameters. 
            :widths: 15 10 10 10
            :header-rows: 1

            * - Parameter
              - Value
              - Bounds
              - Usage
            * - **General**
              - 
              - 
              - 
            * - dt
              - 0.25
              - None
              - Always
            * - tmax
              - 120
              - None
              - Always
            * - dose_tolerance
              - 0.1
              - None
              - Always
            * - t0
              - 0
              - None
              - Always
            * - field_strength
              - 3
              - None
              - Always
            * - **Injection**
              -
              - 
              - 
            * - agent
              - 'gadoxetate'
              - None
              - Always
            * - weight
              - 70
              - None
              - Always
            * - dose
              - 0.0125
              - None
              - Always
            * - rate
              - 1
              - None
              - Always
            * - BAT
              - 60
              - [0, inf]
              - Always
            * - **Sequence**
              -
              - 
              - 
            * - TS
              - 0
              - None
              - Always
            * - TR
              - 0.005
              - None
              - sequence in ['SS', 'SSI']
            * - FA
              - 15
              - None
              - sequence in ['SR', 'SS', 'SSI']
            * - TC
              - 0.1
              - None
              - sequence == 'SR'
            * - TF
              - 0
              - None
              - sequence == 'SSI'
            * - **Aorta**
              -
              - 
              - 

            * - CO
              - 100
              - [0, 300]
              - Always
            * - Thl
              - 10
              - [0, 30]
              - Always
            * - Dhl
              - 0.2
              - [0.05, 0.95]
              - heartlung in ['pfcomp', 'chain']
            * - To
              - 20
              - [0, 60]
              - Always
            * - Eo
              - 0.15
              - [0, 0.5]
              - organs == '2cxm'
            * - Toe
              - 120
              - [0, 800]
              - organs == '2cxm'
            * - Eb
              - 0.05
              - [0.01, 0.15]
              - Always
            * - R10
              - 0.7
              - None
              - Always
            * - S0
              - 1
              - None
              - Always


    Example:

        Use the model to fit minipig aorta data with inflow 
        correction:

    .. plot::
        :include-source:
        :context: close-figs

        >>> import numpy as np
        >>> import pydmr
        >>> import dcmri as dc

        Read the dataset:

        >>> datafile = dc.fetch('minipig_renal_fibrosis')
        >>> data = pydmr.read(datafile, 'nest')
        >>> rois, pars = data['rois']['Pig']['Test'], data['pars']['Pig']['Test']

        Initialize the tissue:

        >>> aorta = dc.Aorta(
        ...     sequence='SSI',
        ...     heartlung='chain',
        ...     organs='comp',
        ...     field_strength=pars['B0'],
        ...     t0=15, 
        ...     agent="gadoterate",
        ...     weight=pars['weight'],
        ...     dose=pars['dose'],
        ...     rate=pars['rate'],
        ...     TR=pars['TR'],
        ...     FA=pars['FA'],
        ...     TS=pars['TS'],
        ...     CO=60, 
        ...     R10=1/dc.T1(pars['B0'], 'blood'),
        ... )

        Create an array of time points:

        >>> time = pars['TS'] * np.arange(len(rois['Aorta']))

        Train the system to the data:

        >>> aorta.train(time, rois['Aorta'])

        Plot the reconstructed signals and concentrations:

        >>> aorta.plot(time, rois['Aorta'])

        Print the model parameters:

        >>> aorta.print_params(round_to=4)

    """

    def __init__(
            self, 
            organs='comp', 
            heartlung='pfcomp', 
            sequence='SS', 
            **params,
        ):

        # Check configuration
        if organs not in ['comp','2cxm']:
            raise ValueError(
                f"{organs} is not a valid model for the organs. "
                "Current options are 'comp' and '2cxm'."
            )
        if heartlung not in ['pfcomp', 'chain']:
            raise ValueError(
                f"{heartlung} is not a valid heart-lung system. "
                "Current options are 'pfcomp' and 'chain'."
            )
        if sequence not in ['SS', 'SR', 'SSI', 'lin']:
            raise ValueError(
                f"Sequence {sequence} is not available."
            )

        # Set configuration
        self.sequence = sequence
        self.organs = organs
        self.heartlung = heartlung

        # Set defaults
        self._set_defaults(**params)
        # For SSI, S0 needs to be free because TF affects the baseline
        if self.sequence == 'SSI':
            self.free['S0'] = [0, np.inf]
        
    def _params(self):
        return PARAMS
    
    def _model_pars(self):

        # General
        pars = ['dt', 'tmax', 'dose_tolerance', 't0', 'field_strength']

        # Injection
        pars += ['agent', 'weight', 'dose', 'rate', 'BAT']

        # Sequence
        pars += ['TS']
        if self.sequence == 'SR':
            pars += ['TC', 'FA']
        elif self.sequence=='SS':
            pars += ['TR', 'FA']
        elif self.sequence=='SSI':
            pars += ['TF', 'TR', 'FA']

        # Aorta
        pars += ['CO', 'Thl', 'To', 'Eb', 'R10', 'S0']
        if self.heartlung == 'pfcomp':
            pars += ['Dhl']
        elif self.heartlung == 'chain':
            pars += ['Dhl']
        if self.organs=='2cxm':
            pars += ['Toe', 'Eo']

        return pars


    def _par_values(self, export=False):

        if export:
            discard = [
                'dt', 'tmax', 't0', 'agent', 'weight', 'dose', 
                'rate', 'field_strength', 'dose_tolerance', 'R10', 
                'TS' , 'TC', 'TR', 'FA',
            ]
            pars = self._par_values()
            return {p: pars[p] for p in pars if p not in discard}
        
        pars = self._model_pars()
        p = {par: getattr(self, par) for par in pars}

        return p
    

    def conc(self):
        """Aorta blood concentration

        Args:
            t (array-like): Time points of the concentration (sec)

        Returns:
            numpy.ndarray: Concentration in M
        """
        if self.organs == 'comp':
            organs = ['comp', (self.To,)]
        elif self.organs=='2cxm':
            organs = ['2cxm', ([self.To, self.Toe], self.Eo)]
        if self.heartlung == 'comp':
            heartlung = ['comp', (self.Thl,)]
        elif self.heartlung == 'pfcomp':
            heartlung = ['pfcomp', (self.Thl, self.Dhl)]
        elif self.heartlung == 'chain':
            heartlung = ['chain', (self.Thl, self.Dhl)]
        self.t = np.arange(0, self.tmax, self.dt)
        conc = lib.ca_conc(self.agent)
        Ji = lib.ca_injection(
            self.t, self.weight, conc, self.dose, self.rate, self.BAT
        )
        Jb = pk_aorta.flux_aorta(
            Ji, E=self.Eb, heartlung=heartlung, organs=organs, 
            dt=self.dt, tol=self.dose_tolerance,
        )
        self.ca = Jb/self.CO
        return self.t, self.ca


    def relax(self):
        """Aorta longitudinal relation rate

        Args:
            t (array-like): Time points of the concentration (sec)

        Returns:
            numpy.ndarray: Concentration in M
        """
        t, cb = self.conc()
        rb = lib.relaxivity(self.field_strength, 'blood', self.agent)
        return t, self.R10 + rb*cb


    def predict(self, xdata) -> np.ndarray:
        tacq = xdata[1]-xdata[0]
        self.tmax = max(xdata)+tacq+self.dt
        if self.TS is not None:
            self.tmax += self.TS
        t, R1 = self.relax()
        if self.sequence == 'SR':
            signal = sig.signal_free(self.S0, R1, self.TC, self.FA)
        elif self.sequence=='SS':
            signal = sig.signal_ss(self.S0, R1, self.TR, self.FA)
        elif self.sequence=='SSI':
            signal = sig.signal_spgr(self.S0, R1, self.TF, self.TR, self.FA, n0=1)
        elif self.sequence == 'lin':
            signal = sig.signal_lin(self.S0, R1)
        return utils.sample(xdata, t, signal, self.TS)


    def train(self, xdata, ydata, **kwargs):
        if self.sequence == 'SR':
            Sref = sig.signal_free(1, self.R10, self.TC, self.FA)
        elif self.sequence=='SS':
            Sref = sig.signal_ss(1, self.R10, self.TR, self.FA)
        elif self.sequence=='SSI':
            Sref = sig.signal_spgr(1, self.R10, self.TF, self.TR, self.FA, n0=1)
        n0 = max([np.sum(xdata < self.t0), 1])
        self.S0 = np.mean(ydata[:n0]) / Sref
        self.BAT = xdata[np.argmax(ydata)] - self.Thl
        return ui.train(self, xdata, ydata, **kwargs)
    

    def plot(self, xdata: np.ndarray, ydata: np.ndarray,
             ref=None, xlim=None, fname=None, show=True):
        aif = self.predict(xdata)
        t, cb = self.conc()
        if xlim is None:
            xlim = [np.amin(xdata), np.amax(xdata)]
        fig, (ax0, ax1) = plt.subplots(1, 2, figsize=(12, 5))
        ax0.set_title('Prediction of the MRI signals.')
        ax0.plot(xdata/60, ydata, color='lightcoral', marker='o', label='Data')
        ax0.plot(xdata/60, aif, color='darkred', linestyle='-', linewidth=3.0, label='Prediction')
        ax0.set_xlabel('Time (min)')
        ax0.set_ylabel('MRI signal (a.u.)')
        ax0.legend()
        ax1.set_title('Prediction of the concentrations.')
        if ref is not None:
            ax1.plot(ref['t']/60, 1000*ref['cb'], 'ro', label='Ground truth')
        ax1.plot(t/60, 1000*cb, color='darkred', linestyle='-', label='Prediction')
        ax1.set_xlabel('Time (min)')
        ax1.set_ylabel('Blood concentration (mM)')
        ax1.legend()
        if fname is not None:
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()


PARAMS = {
    # Prediction and training
    'dt': {
        'init': 0.25,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Forward model time step',
        'unit': 'sec',
    },
    'tmax': {
        'init': 120,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Maximum acquisition time',
        'unit': 'sec',
    },
    'dose_tolerance': {
        'init': 0.1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Dose tolerance',
        'unit': '',
    },
    't0': {
        'init': 0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Baseline duration',
        'unit': 'sec',
    },
    'field_strength': {
        'init': 3.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Magnetic field strength',
        'unit': 'T',
    },

    # Injection
    'agent': {
        'init': 'gadoterate',
        'default_free': False,
        'bounds': None,
        'name': 'Contrast agent',
        'unit': None,
    },
    'weight': {
        'init': 70,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Subject weight',
        'unit': 'kg',
    },
    'dose': {
        'init': lib.ca_std_dose('gadoterate'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Contrast agent dose',
        'unit': 'mL/kg',
    },
    'rate': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Contrast agent injection rate',
        'unit': 'mL/sec',
    },

    # Sequence
    'TR': {
        'init': 0.005,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Repetition time',
        'unit': 'sec',
        'pixel_par': False,
    },
    'FA': {
        'init': 15,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Flip angle',
        'unit': 'deg',
        'pixel_par': False,
    },
    'TC': {
        'init': 0.2,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Time to k-space center',
        'unit': 'sec',
        'pixel_par': False,
    },
    'TS': {
        'init': 0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Sampling time',
        'unit': 'sec',
        'pixel_par': False,
    },
    'TF': {
        'init': 0.50,
        'default_free': True,
        'bounds': [0, 2],
        'name': 'Inflow time',
        'unit': 'sec',
        'pixel_par': False,
    },

    # Aorta
    'BAT': {
        'init': 60,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'Bolus arrival time',
        'unit': 'sec',
    },
    'CO': {
        'init': 100,
        'default_free': True,
        'bounds': [0, 300],
        'name': 'Cardiac output',
        'unit': 'mL/sec',
    },
    'Thl': {
        'init': 10,
        'default_free': True,
        'bounds': [0, 30],
        'name': 'Heart-lung mean transit time',
        'unit': 'sec',
    },
    'Dhl': {
        'init': 0.2,
        'default_free': True,
        'bounds': [0.05, 0.95],
        'name': 'Heart-lung dispersion',
        'unit': '',
    },
    'To': {
        'init': 20,
        'default_free': True,
        'bounds': [0, 60],
        'name': 'Organs blood mean transit time',
        'unit': 'sec',
    },
    'Eo': {
        'init': 0.15,
        'default_free': True,
        'bounds': [0, 0.5],
        'name': 'Organs extraction fraction',
        'unit': '',
    },
    'Toe': {
        'init': 120,
        'default_free': True,
        'bounds': [0, 800],
        'name': 'Organs extravascular mean transit time',
        'unit': 'sec',
    },
    'Eb': {
        'init': 0.05,
        'default_free': True,
        'bounds': [0.01, 0.15],
        'name': 'Body extraction fraction',
        'unit': '',
    },
    'R10': {
        'init': 0.7,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Precontrast R1',
        'unit': 'Hz',
        'pixel_par': False,
    },
    'S0': {
        'init': 1.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Signal scaling factor',
        'unit': 'a.u.',
        'pixel_par': True,
    },

}



import os
import shutil
import numpy as np
import dcmri as dc



# Debugging mode
VERBOSE = 1
SHOW = True

VERBOSE = 0
SHOW = False




def test_ui_tissue():

    # Create data
    time, aif, roi, gt = dc.fake_tissue()

    params = {
        'kinetics': '2CX',
        'water_exchange': 'FF',
        'aif': aif,
        'dt': time[1],
        # 'ca': gt['cb'],
        # 't': gt['t'],
        'r1': dc.relaxivity(3,'blood','gadodiamide'),
        'TR': 0.005,
        'FA': 15,
        'n0': 10,
        'R10a': 1/dc.T1(3.0,'blood'),
        'R10': 1/dc.T1(3.0,'muscle'),
    }

    # Train model and check results
    model = dc.Tissue(**params)
    model.train(time, roi)
    model.plot(time, roi, ref=gt, show=SHOW)
    pars = model.export_params()
    cost0 = model.cost(time, roi)
    assert cost0 < 1 # perc
    assert np.abs(pars['Fp'][1]-gt['Fp']) < 0.1*gt['Fp']
    assert np.abs(pars['vp'][1]-gt['vp']) < 0.1*gt['vp']

    # Try again fixing up to the wrong range
    # Check that the fit is less good
    # And that the value runs against the upper bound
    model = dc.Tissue(vb=0.001, **params)
    model.set_free(vb = [0., 0.01])
    model.train(time, roi)
    pars = model.export_params()
    assert np.round(pars['vb'][1],3)==0.01
    assert model.cost(time, roi) > cost0

    params['water_exchange'] = 'NN'
    model = dc.Tissue(**params)
    model.train(time, roi)
    assert model.cost(time, roi) < 1.5

    params['water_exchange'] = 'RR'
    model = dc.Tissue(**params)
    model.train(time, roi, xtol=1e-1)
    assert model.cost(time, roi) < 1
    assert model.params('PSe') > 1 # fast exchange
    assert model.params('PSc') > 1

    # Loop over all models
    # cost = {'U':20, 'NX':10, 'FX':20, 'WV':10, 'HFU':20, 'HF':6, '2CU':20, 
    #         '2CX':2}
    for k in ['U', 'NX', 'FX', 'WV', 'HFU', 'HF', '2CU', '2CX']:
        for e in ['R','F','N']:
            for c in ['R','F','N']:
                params['water_exchange'] = e+c
                params['kinetics'] = k
                model = dc.Tissue(**params)
                model.train(time, roi, xtol=1e-1)
                #assert model.cost(time, roi) < cost[k]

    # Display last result
    model.plot(time, roi, ref=gt, show=SHOW)
    


def test_ui_tissue_array():

    # Create data
    n=8
    time, signal, aif, gt = dc.fake_brain(n=n, verbose=VERBOSE)

    with np.errstate(divide='ignore'):
        R10 = np.where(gt['T1']==0, 0, 1/gt['T1'])

    # Define model parameters
    params = {
        'kinetics': '2CX',
        'water_exchange': 'FF',
        'aif': aif,
        'dt': time[1],
        'r1': dc.relaxivity(3, 'blood', 'gadodiamide'),
        'TR': 0.005,
        'FA': 15,
        'n0': 10,
        'R10a': 1/dc.T1(3.0,'blood'),
        'R10': R10,
        'parallel': False,
        'verbose': VERBOSE,
    }

    # Train array
    image = dc.TissueArray((n,n), **params)
    image.train(time, signal, xtol=1e-4)
    pars = image.export_params()
    #image.save(path=tmp(), filename='TissueArray')
    #image = dc.TissueArray().load(filename=file)

    # Plot array
    roi = dc.shepp_logan(n=n)
    vmin = {'S0':0, 'Fb':0, 'vb':0, 'PS':0, 'vi':0}
    vmax = {'S0':np.amax(gt['S0']), 'Fb':0.02, 'vb':0.2, 'PS':0.003, 'vi':0.5}
    image.plot(time, signal, vmax=vmax, ref=gt, show=SHOW)
    image.plot_params(roi=roi, ref=gt, vmin=vmin, vmax=vmax, show=SHOW)
    image.plot_fit(time, signal, ref=gt, roi=roi, show=SHOW,
#       hist_kwargs = {'bins':100, 'range':[0,10]},
   )

    # Compare against curve fit in one pixel.
    loc = dc.shepp_logan(n=n)['gray matter'] == 1
    loc = dc.shepp_logan(n=n)['bone'] == 1
    signal_loc = signal[loc,:][0,:]
    params['FA'] = 15
    params['R10'] = R10[loc][0]
    params.pop('parallel')
    params.pop('verbose')
    curve = dc.Tissue(**params)
    curve.train(time, signal_loc, xtol=1e-4)
    curve.plot(time, signal_loc, show=SHOW)

    nrmsc = curve.cost(time, signal_loc)
    nrmsa = image.cost(time, signal)[loc][0]
    assert np.abs(nrmsc-nrmsa) <= 0.1*nrmsc
    cp = curve.params('S0','Fb','vb','PS','vi')
    ip = image.params('S0','Fb','vb','PS','vi')
    for i in range(len(cp)):
        assert np.abs(cp[i] - ip[i][loc][0]) <= 0.1*cp[i]






if __name__ == "__main__":


    # test_ui_tissue()
    test_ui_tissue_array()

    print('All ui_tissue tests passed!!')


import numpy as np
import dcmri as dc


def test_aif_parker():

    t = np.arange(0, 6*60, 1)
    ca = dc.aif_parker(t)

    # Test that this generates values in the right range
    assert np.round(1000*np.amax(ca)) == 6

    # Add a delay and check that this produces the same maximum
    ca = dc.aif_parker(t, BAT=60)
    assert np.round(1000*np.amax(ca)) == 6

    # Try with list as input
    ca = dc.aif_parker([50, 100, 150])
    assert np.array_equal(np.round(1000*ca), [1, 1, 1]) 

    # Or just a single variable
    ca = dc.aif_parker(100)
    assert 1000*ca == 0.7929118932243691

    # Check that an error message is generated if BAT is not a scalar
    try:
        ca = dc.aif_parker(t, BAT=[60,120])
    except: 
        assert True
    else:
        assert False


def test_aif_tristan_rat():
    
    t = np.arange(0, 6*60, 1)
    ca = dc.aif_tristan_rat(t)
    assert np.round(1000*np.amax(ca), 1) == 0.3



if __name__ == "__main__":

    test_aif_parker()
    test_aif_tristan_rat()

    print('All pk_lib tests passed!!')



import numpy as np
import dcmri as dc

def test_params_kidney():
    dc.params_kidney('2CF')

def test_conc_kidney_cm():
    t = np.arange(0, 300, 1.5)
    ca = dc.aif_parker(t, BAT=20)
    params = 0.03, 0.15, 0.8, 4, 10, 60, 60, 30, 30
    Cc, Cm = dc.conc_kidney_cm(ca, *params, t=t, kinetics='7C', sum=True)
    assert round(Cm[100], 4) == 0.0003
    Cc, Cm = dc.conc_kidney_cm(ca, *params, t=t, kinetics='7C', sum=False)
    assert round(Cm[2,100], 5) == 7e-5
    try:
        dc.conc_kidney_cm(ca, kinetics='X')
    except:
        assert True
    else:
        assert False


def test_conc_kidney():
    dt = 1.5
    t = dt*np.arange(20)
    ca = np.ones(20)
    Fp, vp, Ft, Tt = 0.01, 0.2, 0.005, 120
    C = dc.conc_kidney(ca, Fp, vp, Ft, Tt, dt=dt, sum=True, kinetics='2CF')
    assert round(C[10], 1) == 0.1
    C = dc.conc_kidney(ca, Fp, vp, Ft, Tt, dt=dt, sum=False, kinetics='2CF')
    assert round(C[1,10], 2) == 0.02
    C = dc.conc_kidney(ca, vp, Ft, Tt, dt=dt, sum=True, kinetics='HF')
    assert round(C[10], 1) == 0.3
    C = dc.conc_kidney(ca, vp, Ft, Tt, dt=dt, sum=False, kinetics='HF')
    assert round(C[1,10], 2) == 0.07
    Tp = vp/(Fp+Ft)
    h = [1,2,3,4,3,2,1]
    C = dc.conc_kidney(ca, Fp, Tp, Ft, h, dt=dt, sum=True, kinetics='FN')
    assert round(C[10], 1) == 0.2
    C = dc.conc_kidney(ca, Fp, Tp, Ft, h, t=t, sum=True, kinetics='FN')
    assert round(C[10], 1) == 0.1
    C = dc.conc_kidney(ca, Fp, Tp, Ft, h, dt=dt, sum=False, kinetics='FN')
    assert round(C[1,10], 2) == 0.02

    try:
        dc.conc_kidney(ca, kinetics='X')
    except:
        assert True
    else:
        assert False


if __name__ == '__main__':

    test_params_kidney()
    test_conc_kidney()
    test_conc_kidney_cm()

    print('All kidney tests passed!!')


import numpy as np
import dcmri as dc


def test_signal_dsc():
    S0 = 1
    TE = 0
    TR = 0
    R2 = 1
    R1 = 1
    S = dc.signal_dsc(R1, R2, S0, TR, TE)
    assert S==0


def test_signal_t2w():
    S0 = 1
    TE = 0
    R2 = 1
    S = dc.signal_t2w(R2, S0, TE)
    assert S==1

def test_Mz_free():

    # Tests 1 (may duplicate some of Tests 2 below)
    R1 = 1
    TI = 0.1*np.arange(100)
    f = 0.5

    Mz = dc.Mz_free(R1, TI, n0=-1)
    Mz_e = dc.Mz_free(R1, TI, n0=-1, Fw=f, j=f)
    Mz_i = dc.Mz_free(R1, TI, n0=-1, Fw=f, j=-f)

    assert 21 < np.linalg.norm(Mz + Mz_e + Mz_i) < 22

    R1 = [1,2]
    v = [0.3, 0.7]
    PS = 0.1
    Fw = [[f, PS], [PS, 0]]
    Mz = dc.Mz_free(R1, TI, v, Fw, n0=-1, j=[f, 0])

    assert 7 < np.linalg.norm(Mz) < 8

    TI = 0.5
    nt = 1000
    t = 0.1*np.arange(nt)
    R1 = np.stack((1-t/np.amax(t), np.ones(nt)))
    j = np.stack((f*np.ones(nt), np.zeros(nt)))
    Mz = dc.Mz_free(R1, TI, v, Fw, n0=-1, j=j)

    assert 5 < np.linalg.norm(Mz) < 6

    TI = 0.1*np.arange(10)
    Mzi = dc.Mz_free(R1, TI, v, Fw, n0=-1, j=j)

    assert np.linalg.norm(Mzi[:,:,5]-Mz) == 0

    # Tests 2
    R1 = 1
    T = [1,2]
    S = dc.Mz_free(R1, T)
    assert 0.6 < S[0] < 0.7
    R1 = [1,1]
    S = dc.Mz_free(R1, T)
    assert 0.6 < S[0,0] < 0.7
    R1 = 1
    T = 1
    S = dc.Mz_free(R1, T)
    assert 0.6 < S < 0.7
    R1 = [1,1]
    S = dc.Mz_free(R1, T)
    assert 0.6 < S[0] < 0.7
    v = [0.2, 0.3]
    S = dc.Mz_free(R1, T, v)
    assert 0.1 < S[0] < 0.2
    R1 = np.ones((2,3))
    j = np.ones((2,3))
    S = dc.Mz_free(R1, T, v, j=j)
    assert 0.7 < S[0,0] < 0.8
    S = dc.Mz_free(R1, T, v)
    assert 0.1 < S[0,0] < 0.2
    S = dc.Mz_free(R1, T, v, n0=[0,0])
    assert 0.1 < S[0,0] < 0.2
    S = dc.Mz_free(R1, T, v, n0=np.zeros((2,3)))
    assert 0.1 < S[0,0] < 0.2

def test_Mz_ss():

    # Functional tests

    # Steady state magnetization with different inflows
    R1 = 1
    FA = 12
    TR = 0.005
    f = 0.5
    v = 0.7

    # Check that inflow with steady-state magnetization 
    # is the same as no inflow
    m_c = dc.Mz_ss(R1, TR, FA, v)/v
    m_f = dc.Mz_ss(R1, TR, FA, v, Fw=f, j=f*m_c)/v
    assert np.abs(m_f-m_c) < 0.01*np.abs(m_c)

    # Repeat for a two-compartment system:
    R1 = [0.5, 1.5]
    v = [0.3, 0.6]
    PS = 0.1

    Fw = [[0, PS], [PS, 0]]
    M_c = dc.Mz_ss(R1, TR, FA, v, Fw)

    Fw = [[f, PS], [PS, 0]]
    m_c = M_c[0]/v[0]
    m_f = dc.Mz_ss(R1, TR, FA, v, Fw, j=[f*m_c, 0])[0]/v[0]

    assert np.abs(m_c-m_f) < 0.01*np.abs(m_c)

    # Check fast-exchange limit
    v = [0.3, 0.5]
    M_c = dc.Mz_ss(R1, TR, FA, v, 1e9)
    M_c_fex = dc.Mz_ss(R1, TR, FA, v, np.inf)
    assert np.linalg.norm(M_c-M_c_fex) < 1e-3*np.linalg.norm(M_c_fex)

    # Check no-exchange limit
    v = [0.3, 0.5]
    M_c = dc.Mz_ss(R1, TR, FA, v, 1e-9)
    M_c_nex = dc.Mz_ss(R1, TR, FA, v, 0)
    assert np.linalg.norm(M_c-M_c_nex) < 1e-6*np.linalg.norm(M_c_nex)

    # Check exceptions
    Fw = [[0, np.inf], [np.inf, 0]]
    M_c_fex1 = dc.Mz_ss(R1, TR, FA, v, Fw)
    M_c_fex2 = dc.Mz_ss(R1, TR, FA, v, np.inf)
    assert np.linalg.norm(M_c_fex1-M_c_fex2) < 1e-9*np.linalg.norm(M_c_fex2)

    Fw = [[0, 0], [0, 0]]
    M_c_nex1 = dc.Mz_ss(R1, TR, FA, v, Fw)
    M_c_nex2 = dc.Mz_ss(R1, TR, FA, v, 0)
    assert np.linalg.norm(M_c_nex1-M_c_nex2) < 1e-9*np.linalg.norm(M_c_nex2)

    try:
        Fw = [[0, np.inf], [0, 0]]
        M_c_fex1 = dc.Mz_ss(R1, TR, FA, v, Fw)  
    except:
        assert True
    else:
        assert False   

    # Check all cases
    R1 = 1
    TR = 0.005
    FA = 15
    S = dc.Mz_ss(R1, TR, FA)
    assert 0.1 < S < 0.2
    S = dc.Mz_ss(0, TR, FA)
    assert S==0
    R1 = [1,1]
    S = dc.Mz_ss(R1, TR, FA)
    assert 0.1 < S[0] < 0.2
    v = [0.2, 0.3]
    S = dc.Mz_ss(R1, TR, FA, v)
    assert 0.02 < S[0] < 0.03
    S = dc.Mz_ss(R1, TR, FA, v, Fw=0.1)
    assert 0.02 < S[0] < 0.03
    R1 = np.ones((2,3))
    j = np.ones((2,3))
    S = dc.Mz_ss(R1, TR, FA, v, j=j)
    assert 0.02 < S[0,0] < 0.03
    S = dc.Mz_ss(R1, TR, FA, v)
    assert 0.02 < S[0,0] < 0.03


def test_Mz_spgr():

    # Functional test
    FA = 12
    TR = 0.005
    TI = np.linspace(0,3,100)
    TP = 0

    R1 = [1, 0.5]
    v = [0.3, 0.7]
    f = 0.5
    PS = 0.1
    Fw = [[f, PS], [PS, 0]]
    Mspgr = dc.Mz_spgr(R1, TI, TR, FA, TP, v, Fw, j=[f, 0], n0=-1) 
    Mss = dc.Mz_ss(R1, TR, FA, v, Fw, j=[f, 0])

    # Check that SPGR converges to steady state
    assert np.linalg.norm(Mspgr[:,-1]-Mss) < 1e-4*np.linalg.norm(Mss)

    # Test cases
    R1 = 1
    T = [1, 2]
    TR = 0.005
    FA = 15
    TP= 0
    S = dc.Mz_spgr(R1, T, TR, FA, TP)
    assert 0.1 < S[0] < 0.2
    R1 = [1,1]
    S = dc.Mz_spgr(R1, T, TR, FA, TP)
    assert 0.1 < S[0,0] < 0.2
    R1 = 1
    T = 1
    S = dc.Mz_spgr(R1, T, TR, FA, TP)
    assert 0.1 < S < 0.2
    R1 = [1,1]
    S = dc.Mz_spgr(R1, T, TR, FA, TP)
    assert 0.1 < S[0] < 0.2
    v = [0.2, 0.3]
    S = dc.Mz_spgr(R1, T, TR, FA, TP, v)
    assert 0.02 < S[0] < 0.03
    R1 = np.ones((2,3))
    j = np.ones((2,3))
    S = dc.Mz_spgr(R1, T, TR, FA, TP, v, j=j)
    assert 0.02 < S[0,0] < 0.03
    S = dc.Mz_spgr(R1, T, TR, FA, TP, v)
    assert 0.02 < S[0,0] < 0.03

def test_signal_ss():
    R1 = 1
    S0 = 1
    TR = 1
    FA = 0
    S = dc.signal_ss(S0, R1, TR, FA)
    assert S==0
    v = [0.5, 0.5]
    try:
        S = dc.signal_ss(S0, R1, TR, FA, v=v)
    except:
        assert True
    else:
        assert False
    R1 = [1,1]
    S = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=np.inf)
    assert S==0
    S = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=0)
    assert S==0
    Fw = [[0,1],[1,0]]
    S = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=Fw)
    assert S==0
    R1 = np.ones((2,10))
    S = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=np.inf)
    assert 0==np.linalg.norm(S)
    S = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=0)
    assert 0==np.linalg.norm(S)
    R1 = [1,1]
    S = dc.signal_ss(S0, R1, TR, FA)
    assert 0==np.linalg.norm(S)

    # Calibrate on other R10 reference
    R1 = 1
    S0 = 5
    TR = 1
    FA = 45
    S = dc.signal_ss(S0, R1, TR, FA, R10=R1)
    assert np.round(S)==5
    R1 = [1,1]
    v = [0.5, 0.5]
    S = dc.signal_ss(S0, R1, TR, FA, v=v, R10=R1, Fw=np.inf)
    assert np.round(S)==5

    # Check exceptions
    try:
        S = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=[1,1])
    except:
        assert True
    else:
        assert False

    try:
        S = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=np.ones((3,3)))
    except:
        assert True
    else:
        assert False

    v = [0.1, 0.4, 0.5]
    try:
        S = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=0)
    except:
        assert True
    else:
        assert False

    R1 = [1,1]
    try:
        S = dc.signal_ss(S0, R1, TR, FA, v=v)
    except:
        assert True
    else:
        assert False

    # Check boundary regimes with flow
    n = 100
    R1 = np.ones((2,n))
    j = np.ones((2,n))
    v = [0.5, 0.5]
    S0 = 1
    TR = 1
    FA = 10

    # No water exchange
    zero = 1e-6
    Fw = [[0.1,zero],[zero,1]]
    S1 = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=Fw, j=j)
    Fw = [[0.1,0],[0,1]]
    S2 = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=Fw, j=j)
    assert np.linalg.norm(S1-S2) < 1e-6*np.linalg.norm(S2)

    # Fast water exchange
    inf = 1e+6
    Fw = np.array([[0.1,inf],[inf,1]])
    S1 = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=Fw, j=j)
    Fw = np.array([[0.1,np.inf],[np.inf,1]])
    S2 = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=Fw, j=j)
    assert np.linalg.norm(S1-S2) < 1e-6*np.linalg.norm(S2)
    Fw = [[0.1,inf],[np.inf,1]]
    try:
        S2 = dc.signal_ss(S0, R1, TR, FA, v=v, Fw=Fw, j=j)
    except:
        assert True
    else:
        assert False
    # assert np.linalg.norm(S1-S2) < 1e-6*np.linalg.norm(S2)


def test_signal_spgr():
    R1 = 1
    S0 = 1
    TR = 1
    FA = 0
    TC = 1
    S = dc.signal_spgr(S0, R1, TC, TR, FA, n0=0)
    assert S==0
   
    v = [0.5, 0.5]
    try:
        S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, n0=0)
    except:
        assert True
    else:
        assert False
    R1 = [1,1]
    S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, n0=0)
    assert S==0
    S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, Fw=0, n0=0)
    assert S==0
    Fw = [[0,1],[1,0]]
    S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, Fw=Fw, n0=0)
    assert S==0
    R1 = np.ones((2,10))
    S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, n0=0)
    assert 0==np.linalg.norm(S)
    S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, Fw=0, n0=0)
    assert 0==np.linalg.norm(S)
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP=10, v=v, Fw=0, n0=0)
    assert 0==np.linalg.norm(S)
    S = dc.signal_spgr(S0, [1,1], TC, TR, FA, TP=10, v=v, Fw=0, n0=0)
    assert 0==np.linalg.norm(S)

    # Check exceptions
    try:
        S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, Fw=[1,1], n0=0)
    except:
        assert True
    else:
        assert False

    v = [0.1, 0.4, 0.5]
    try:
        S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, Fw=0, n0=0)
    except:
        assert True
    else:
        assert False

    R1 = [1,1]
    try:
        S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, n0=0)
    except:
        assert True
    else:
        assert False

    # Calibrate on other R10 reference
    R1 = 1
    S0 = 5
    TR = 1
    FA = 45
    S = dc.signal_spgr(S0, R1, TC, TR, FA, R10=R1, n0=0)
    assert np.round(S)==5
    R1 = [1,1]
    v = [0.5, 0.5]
    S = dc.signal_spgr(S0, R1, TC, TR, FA, v=v, R10=R1, n0=0)
    assert np.round(S)==5

    # With preparation
    R1 = 1
    S0 = 1
    TR = 1
    FA = 0
    TC = 1
    TP = 0
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP)
    assert S==0
    v = [0.5, 0.5]
    try:
        S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v)
    except:
        assert True
    else:
        assert False
    R1 = [1,1]
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v)
    assert S==0
    R1 = [1,1]
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP)
    assert S[0]==0
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v, Fw=0)
    assert S==0
    Fw = [[0,1],[1,0]]
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v, Fw=Fw)
    assert S==0
    R1 = np.ones((2,10))
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v)
    assert 0==np.linalg.norm(S)
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v, Fw=0)
    assert 0==np.linalg.norm(S)

    # Check exceptions
    try:
        S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v, Fw=[1,1])
    except:
        assert True
    else:
        assert False

    try:
        S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v, Fw=np.ones((3,3)))
    except:
        assert True
    else:
        assert False

    v = [0.1, 0.4, 0.5]
    try:
        S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v, Fw=0)
    except:
        assert True
    else:
        assert False

    R1 = [1,1]
    try:
        S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v)
    except:
        assert True
    else:
        assert False

    # Calibrate on other R10 reference
    R1 = 1
    S0 = 5
    TR = 1
    FA = 45
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, R10=R1)
    assert np.round(S)==5
    R1 = [1,1]
    v = [0.5, 0.5]
    S = dc.signal_spgr(S0, R1, TC, TR, FA, TP, v=v, R10=R1)
    assert np.round(S)==5

    # Check convergence to fast exchange


def test_signal_free():
    TC = 0
    R1 = 1
    S0 = 1 
    FA = 45
    S = dc.signal_free(S0, R1, TC, FA, R10=None)
    assert S==0
    TC = 1
    S0 = 0
    S = dc.signal_free(S0, R1, TC, FA, R10=1)
    assert S==0
    R1 = [1,1]
    S0 = 1
    S = dc.signal_free(S0, R1, TC, FA)
    assert 0.4 < S[0] < 0.5
    S = dc.signal_free(S0, R1, TC, FA, v=[0.2, 0.3])
    assert 0.2 < S < 0.3
    R1 = np.ones((2,3))
    S = dc.signal_free(S0, R1, TC, FA, v=[0.2, 0.3])
    assert 0.2 < S[0] < 0.3


def test_signal_lin():
    R1 = 2
    S0 = 3
    assert 6 == dc.signal_lin(S0, R1)


def test_conc_t2w():
    S = np.ones(10)
    TE = np.inf
    C = dc.conc_t2w(S, TE, r2=1, n0=1)
    assert 0 == np.linalg.norm(C)
    
def test_conc_ss():
    S = np.ones(10)
    TR = 1
    FA = 45
    T10 = np.inf
    C = dc.conc_ss(S, TR, FA, T10, r1=1, n0=1)
    assert 0 == np.linalg.norm(C)

    # Data without solution
    S = [1,2,0]
    T10 = 1
    C = dc.conc_ss(S, TR, FA, T10, r1=1, n0=1)
    assert C[-1] == -1

def test_conc_src():
    S = np.ones(10)
    TC = 1
    T10 = np.inf
    C = dc.conc_src(S, TC, T10, r1=1, n0=1)
    assert 0 == np.linalg.norm(C)

def test_conc_lin():
    S = np.ones(10)
    T10 = 1
    C = dc.conc_lin(S, T10, r1=1, n0=1)
    assert np.linalg.norm(C)==0

def test_signal():
    S = dc.signal('SRC', 1, 2, TC=0.1, FA=10)
    assert 0.03 < S < 0.04
    S = dc.signal('SR', 1, 2, TC=0.1, FA=10, TR=0.005)
    assert 0.02 < S < 0.03
    S = dc.signal('SS', 1, 2, FA=10, TR=0.005)
    assert 0.08 < S < 0.09
    try:
        S = dc.signal('X', 1, 2)
    except:
        assert True
    else:
        assert False


if __name__ == "__main__":

    test_signal()
    test_signal_dsc()
    test_signal_t2w()
    test_Mz_free()
    test_Mz_ss()
    test_Mz_spgr()
    test_signal_ss()
    test_signal_spgr()
    test_signal_free()
    test_signal_lin()

    test_conc_t2w()
    test_conc_ss()
    test_conc_src()
    test_conc_lin()

    print('All sig tests passing!')


import numpy as np

import dcmri.pk as pk



def aif_parker(t, BAT: float = 0.0) -> np.ndarray:
    """Population AIF model as defined by `Parker et al (2006) <https://onlinelibrary.wiley.com/doi/full/10.1002/mrm.21066>`_

    Args:
        t (array_like): time points in units of sec.
        BAT (float, optional): Time in seconds before the bolus arrives. Defaults to 0 sec (no delay).

    Returns:
        np.ndarray: Concentrations in M for each time point in t. If t is a scalar, the return value is a scalar too.

    References:
        Adapted from a contribution by the QBI lab of the University of Manchester to the `OSIPI code repository <https://github.com/OSIPI/DCE-DSC-MRI_CodeCollection>`_.

    Example:

        >>> import numpy as np
        >>> import dcmri as dc

        Create an array of time points covering 20sec in steps of 1.5sec, which rougly corresponds to the first pass of the Paeker AIF:

        >>> t = np.arange(0, 20, 1.5)

        Calculate the Parker AIF at these time points, and output the result in units of mM:

        >>> 1000*dc.aif_parker(t)
        array([0.08038467, 0.23977987, 0.63896354, 1.45093969,
        2.75255937, 4.32881325, 5.6309778 , 6.06793854, 5.45203828,
        4.1540079 , 2.79568217, 1.81335784, 1.29063036, 1.08751679])
    """

    # Check input types
    if not np.isscalar(BAT):
        raise ValueError('BAT must be a scalar')

    # Convert from secs to units used internally (mins)
    t_offset = np.array(t) / 60 - BAT / 60

    # A1/(SD1*sqrt(2*PI)) * exp(-(t_offset-m1)^2/(2*var1))
    # A1 = 0.833, SD1 = 0.055, m1 = 0.171
    gaussian1 = 5.73258 * np.exp(
        -1.0 * (t_offset - 0.17046) * (t_offset - 0.17046) / (2.0 * 0.0563 * 0.0563))

    # A2/(SD2*sqrt(2*PI)) * exp(-(t_offset-m2)^2/(2*var2))
    # A2 = 0.336, SD2 = 0.134, m2 = 0.364
    gaussian2 = 0.997356 * np.exp(
        -1.0 * (t_offset - 0.365) * (t_offset - 0.365) / (2.0 * 0.132 * 0.132))
    # alpha*exp(-beta*t_offset) / (1+exp(-s(t_offset-tau)))
    # alpha = 1.064, beta = 0.166, s = 37.772, tau = 0.482
    sigmoid = 1.050 * np.exp(-0.1685 * t_offset) / \
        (1.0 + np.exp(-38.078 * (t_offset - 0.483)))

    pop_aif = gaussian1 + gaussian2 + sigmoid

    return pop_aif / 1000  # convert to M


def aif_tristan_rat(t, BAT=4.6 * 60, duration=30) -> np.ndarray:
    """Population AIF model for rats measured with a standard dose of 
    gadoxetate.

    Args:
        t (array_like): time points in units of sec.
        BAT (float, optional): Time in seconds before the bolus arrives. 
          Defaults to 4.6 min.
        duration (float, optional): Duration of the injection. Defaults to 30s.

    Returns:
        np.ndarray: Blood concentrations in M for each time point in t. If 
        t is a scalar, the return value is a scalar too.

    References:

        - Melillo N, Scotcher D, Kenna JG, Green C, Hines CDG, Laitinen I, 
          et al. Use of In Vivo Imaging and Physiologically-Based Kinetic 
          Modelling to Predict Hepatic Transporter Mediated Drug-Drug 
          Interactions in Rats. 
          `Pharmaceutics 2023;15(3):896 <https://doi.org/10.3390/pharmaceutics15030896>`_.

        - Gunwhy, E. R., & Sourbron, S. (2023). TRISTAN-RAT (v3.0.0). 
          `Zenodo <https://doi.org/10.5281/zenodo.8372595>`_

    Example:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        Create an array of time points over 30 minutes

        >>> t = np.arange(0, 30*60, 0.1)

        Generate the rat input function for these time points:

        >>> cb = dc.aif_tristan_rat(t)

        Plot the result:

        >>> plt.plot(t/60, 1000*cb, 'r-')
        >>> plt.title('TRISTAN rat AIF')
        >>> plt.xlabel('Time (min)')
        >>> plt.ylabel('Blood concentration (mM)')
        >>> plt.show()
    """
    # Constants

    dose = 0.0075   # mmol

    Fb = 2.27 / 60    # mL/sec
    # https://doi.org/10.1021/acs.molpharmaceut.1c00206
    # (Changed from 3.61/60 on 07/03/2022)
    # From Brown the cardiac output of rats is
    # 110.4 mL/min (table 3-1) ~ 6.62L/h
    # From table 3-4, sum of hepatic artery and portal vein
    # blood flow is 17.4% of total cardiac output ~ 1.152 L/h
    # Mass of liver is 9.15g, with density of 1.08 kg/L,
    # therefore ~8.47mL
    #  9.18g refers to the whole liver, i.e. intracellular tissue
    # + extracellular space + blood
    # Dividing 1.152L/h for 8.47mL we obtain ~2.27 mL/min/mL liver
    # Calculation done with values in Table S2 of our article
    # lead to the same results
    Hct = 0.418     # Cremer et al, J Cereb Blood Flow Metab 3, 254-256 (1983)
    VL = 8.47       # mL
    # Scotcher et al 2021, DOI: 10.1021/acs.molpharmaceut.1c00206
    # Supplementary material, Table S2
    GFR = 1.38 / 60   # mL/sec (=1.38mL/min)
    # https://doi.org/10.1152/ajprenal.1985.248.5.F734
    P = 0.172       # mL/sec
    # Estimated from rat repro study data using PBPK model
    # 0.62 L/h
    # Table 3 in Scotcher et al 2021
    # DOI: 10.1021/acs.molpharmaceut.1c00206
    VB = 15.8       # mL
    # 0.06 X BW + 0.77, Assuming body weight (BW) = 250 g
    # Lee and Blaufox. Blood volume in the rat.
    # J Nucl Med. 1985 Jan;26(1):72-6.
    VE = 30         # mL
    # All tissues, including liver.
    # Derived from Supplementary material, Table S2
    # Scotcher et al 2021
    # DOI: 10.1021/acs.molpharmaceut.1c00206
    E = 0.4         # Liver extraction fraction, estimated from TRISTAN data
    # Using a model with free E, then median over population of controls

    # Derived constants
    VP = (1 - Hct) * VB
    Fp = (1 - Hct) * Fb
    K = GFR + E * Fp * VL  # mL/sec

    # Influx in mmol/sec
    J = np.zeros(np.size(t))
    Jmax = dose / duration  # mmol/sec
    J[(t > BAT) & (t < BAT + duration)] = Jmax

    # Model parameters
    TP = VP / (K + P)
    TE = VE / P
    E = P / (K + P)

    Jp = pk.flux_2cxm(J, [TP, TE], E, t)
    cp = Jp / K

    return cp * (1 - Hct)




from scipy.linalg import expm
from scipy.special import i0, i1
import numpy as np
from tqdm import tqdm


def _signal_rice(nu, sigma):
    if sigma==0:
        return nu
    with np.errstate(divide='ignore', over='ignore', invalid='ignore'):
        K = nu**2 / (2*sigma**2)
        arg = K/2
        pref = sigma * np.sqrt(np.pi/2)
        rice_mean = pref * np.exp(-K/2) * ((1+K)*i0(arg) + K*i1(arg))
    # Nan values are points where the distribution is indistinguisable from Gaussian
    return np.where(np.isnan(rice_mean) | np.isinf(rice_mean), nu, rice_mean)



# TODO handle the case where F of one or more compartments is infinite
def _Mz_K(R1, v, Fw):
    if np.isscalar(R1):
        return R1 + Fw/v
    nc = np.size(R1)
    K = np.zeros((nc, nc))
    if np.isscalar(Fw):
        Fw = np.full((nc, nc), Fw) - np.diag(np.full(nc, Fw))  
    elif isinstance(Fw, list):
        Fw = np.array(Fw)
    F = np.sum(Fw, axis=0)
    for c in range(nc):
        K[c, c] = R1[c] + F[c]/v[c]
        for d in range(nc):
            if d != c:
                K[c, d] = -Fw[c, d]/v[d]
    return K


def _Mz_J(R1, v, me, j=None):
    J = np.multiply(np.multiply(R1, v), me)
    J = J if j is None else J + np.multiply(j, me)
    return J


def Mz_free(R1, T, v=1, Fw=0, j=None, n0=0, me=1):
    """Free longitudinal magnetization.

    See section :ref:`basics-relaxation-T1` for more detail.

    Args:
        R1 (array-like): Longitudinal relaxation rates in 1/sec. For a tissue 
          with n compartments, the first dimension of R1 must be n. For a
          single compartment, R1 can be scalar or a 1D time-array.
        T (array-like): duration of free recovery.
        v (array-like, optional): volume fractions of the compartments. For a 
          one-compartment tissue this is a scalar - otherwise it is an 
          array with one value for each compartment. Defaults to 1.
        Fw (array-like, optional): Water flow between the compartments and to 
          the environment, in units of mL/sec/cm3. Generally Fw must be a nxn 
          array, where n is the number of compartments, and the off-diagonal 
          elements Fw[j,i] are the permeability for water moving from 
          compartment i into j. The diagonal elements Fw[i,i] quantify the 
          flow of water from compartment i to outside. For a closed system 
          with equal permeabilities between all compartments, a scalar value 
          for Fw can be provided. Defaults to 0.
        j (array-like, optional): normalized tissue magnetization flux. j has 
          to have the same shape as R1. Defaults to None.
        n0 (array-like, optional): initial relative magnetization at T=0. 
          If this is a scalar, all compartments are assumed to have the same 
          initial magnetization. Defaults to 0.
        me (array-like, optional): equilibrium magnetization of the tissue 
          compartments. If a scalar value is provided, all compartments are 
          assumed to have the same equilibrium magnetization. Defaults to 1.

    Returns:
        np.ndarray: Magnetization in the compartments after a time T.

    Example:

        Magnetization recovery after inversion.

    .. plot::
        :include-source:
        :context: close-figs

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Plot magnetization recovery for the first 10 seconds after an 
        inversion pulse, for a closed tissue with R1 = 1 sec, and for an open 
        tissue with equilibrium inflow and inverted inflow:

        >>> TI = 0.1*np.arange(100)
        >>> R1 = 1
        >>> f = 0.5

        >>> Mz = dc.Mz_free(R1, TI, n0=-1)
        >>> Mz_e = dc.Mz_free(R1, TI, n0=-1, Fw=f, j=f)
        >>> Mz_i = dc.Mz_free(R1, TI, n0=-1, Fw=f, j=-f)

        >>> plt.plot(TI, Mz, label='No flow', linewidth=3)
        >>> plt.plot(TI, Mz_e, label='Equilibrium inflow', linewidth=3)
        >>> plt.plot(TI, Mz_i, label='Inverted inflow', linewidth=3)
        >>> plt.xlabel('Inversion time (sec)')
        >>> plt.ylabel('Magnetization (A/cm)')
        >>> plt.legend()
        >>> plt.show()

        Now consider a two-compartment model, with a central compartment 
        that has in- and outflow, and a peripheral compartment that only 
        exchanges with the central compartment:

        >>> R1 = [1,2]
        >>> v = [0.3, 0.7]
        >>> PS = 0.1
        >>> Fw = [[f, PS], [PS, 0]]
        >>> Mz = dc.Mz_free(R1, TI, v, Fw, n0=-1, j=[f, 0])

        >>> plt.plot(TI, Mz[0,:], label='Central compartment', linewidth=3)
        >>> plt.plot(TI, Mz[1,:], label='Peripheral compartment', linewidth=3)
        >>> plt.xlabel('Inversion time (sec)')
        >>> plt.ylabel('Magnetization (A/cm)')
        >>> plt.legend()
        >>> plt.show()

        In DC-MRI the more usual situation is one where TI is fixed and the 
        relaxation rates are variable due to the effect of a contrast agent. 
        As an illustration, consider the previous result again at TI=500 msec 
        and an R1 that is linearly declining in the central compartment and 
        constant in the peripheral compartment:

        >>> TI = 0.5
        >>> nt = 1000
        >>> t = 0.1*np.arange(nt)
        >>> R1 = np.stack((1-t/np.amax(t), np.ones(nt)))
        >>> j = np.stack((f*np.ones(nt), np.zeros(nt)))
        >>> Mz = dc.Mz_free(R1, TI, v, Fw, n0=-1, j=j)

        >>> plt.plot(t, Mz[0,:], label='Central compartment', linewidth=3)
        >>> plt.plot(t, Mz[1,:], label='Peripheral compartment', linewidth=3)
        >>> plt.xlabel('Time (sec)')
        >>> plt.ylabel('Magnetization (A/cm)')
        >>> plt.legend()
        >>> plt.show()   

        The function allows for R1 and TI to be both variable. Computing the 
        result for 10 different TI values and extracting the result 
        corresponding to TI=0.5 gives again the same result:

        >>> TI = 0.1*np.arange(10)
        >>> Mz = dc.Mz_free(R1, TI, v, Fw, n0=-1, j=j)

        >>> plt.plot(t, Mz[0,:,5], label='Central compartment', linewidth=3)
        >>> plt.plot(t, Mz[1,:,5], label='Peripheral compartment', linewidth=3)
        >>> plt.xlabel('Time (sec)')
        >>> plt.ylabel('Magnetization (A/cm)')
        >>> plt.legend()
        >>> plt.show()      

    """
    if not np.isscalar(T): # Do we need this?
        if np.isscalar(R1):
            Mz = np.empty(np.size(T))
            for k, Tk in enumerate(T):
                Mz[k] = Mz_free(R1, Tk, v, Fw, j, n0, me)
        else:
            Mz = np.empty(np.shape(R1) + (np.size(T), ))
            for k, Tk in enumerate(T):
                Mz[...,k] = Mz_free(R1, Tk, v, Fw, j, n0, me)
        return Mz

    # One compartment
    if np.isscalar(v):
        if np.isscalar(R1):
            K = _Mz_K(R1, v, Fw)
            J = _Mz_J(R1, v, me, j=j)
            E = np.exp(-T*K)
            Kinv = 1/K
            M0 = n0*me*v
            return E*M0 + (1-E)*Kinv*J
        else:
            R1 = np.array(R1)
            nt = np.size(R1)
            M = np.empty(nt)
            for t in range(nt):
                jt = None if j is None else j[t]
                n0t = n0 if np.isscalar(n0) else n0[t]
                M[t] = Mz_free(R1.reshape(-1)[t], T, v, Fw, jt, n0t, me)
            return M
        
    # Multiple compartments  
    else:
        
        n = len(v)
        if np.isscalar(R1):
            raise ValueError(
                'In a tissue with n compartments, R1 must have at least n elements.')
        R1 = np.array(R1)
        if R1.shape[0] != n:
            raise ValueError(
                'In a tissue with n compartments, the first dimension of R1 must have size n.')

        if np.ndim(R1)==1:
            K = _Mz_K(R1, v, Fw)
            J = _Mz_J(R1, v, me, j=j)
            E = expm(-T*K)
            Kinv = np.linalg.inv(K)
            I = np.eye(np.size(R1))
            M0 = np.multiply(np.multiply(n0, me), v)
            return np.dot(E, M0) + np.dot(I-E, np.dot(Kinv, J))

        shape = R1.shape
        R1 = R1.reshape(n, -1)
        if j is not None:
            j = j.reshape(n, -1)
        M = np.zeros(R1.shape)
        for t in range(R1.shape[1]):
            jt = None if j is None else j[:,t]
            if np.ndim(n0)==2:
                n0t = n0[:,t]
            else:
                n0t = n0
            M[:,t] = Mz_free(R1[:,t], T, v, Fw, jt, n0t, me) 
        return M.reshape(shape)  


def Mz_ss(R1, TR, FA, v=1, Fw=0, j=None, me=1) -> np.ndarray:
    """Steady-state longitudinal tissue magnetization.

    See section :ref:`basics-relaxation-T1` for more detail.

    Args:
        R1 (array-like): Longitudinal relaxation rates in 1/sec. For a tissue 
          with n compartments, the first dimension of R1 must be n. For a
          single compartment, R1 can be scalar or a 1D time-array.
        TR (float): repetition time.
        FA (float): flip angle.
        v (array-like, optional): volume fractions of the compartments. For a 
          one-compartment tissue this is a scalar - otherwise it is an 
          array with one value for each compartment. Defaults to 1.
        Fw (array-like, optional): Water flow between the compartments and to 
          the environment, in units of mL/sec/cm3. Generally Fw must be a nxn 
          array, where n is the number of compartments, and the off-diagonal 
          elements Fw[j,i] are the permeability for water moving from 
          compartment i into j. The diagonal elements Fw[i,i] quantify the 
          flow of water from compartment i to outside. For a closed system 
          with equal permeabilities between all compartments, a scalar value 
          for Fw can be provided. Defaults to 0.
        j (array-like, optional): normalized tissue magnetization flux. j has 
          to have the same shape as R1. Defaults to None.
        me (array-like, optional): equilibrium magnetization of the tissue 
          compartments. If a scalar value is provided, all compartments are 
          assumed to have the same equilibrium magnetization. Defaults to 1.

    Returns:
        np.ndarray: Magnetization in the compartments after a time T.

    Example:

        Compute steady-state magnetization with inflow magnetization (mi) 
        ranging from fully inverted to equilibrium. Compare against 
        magnetization of an isolated tissue without flow.

    .. plot::
        :include-source:
        :context: close-figs

        Import packages:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Define constants:

        >>> FA, TR = 12, 0.005
        >>> R1 = 1
        >>> f, v = 0.5, 0.7
        >>> mi = np.linspace(-1, 1, 100)

        Compute magnetization without/with inflow:

        >>> m_c = dc.Mz_ss(R1, TR, FA, v)/v
        >>> m_f = [dc.Mz_ss(R1, TR, FA, v, f, j=f*m)/v for m in mi]

        Plot the results:

        >>> plt.plot(mi, mi*0+m_c, label='No inflow', linewidth=3)
        >>> plt.plot(mi, m_f, label='Inflow', linewidth=3)
        >>> plt.xlabel('Inflow magnetization (A/cm/mL)')
        >>> plt.ylabel('Steady-state magnetization (A/cm/mL)')
        >>> plt.legend()
        >>> plt.show()

        Note the magnetization of the two tissues is the same when the 
        inflow is at the steady-state of the isolated tissue: 

        >>> m_f = dc.Mz_ss(R1, TR, FA, v, f, j=f*m_c)/v
        >>> print(m_f - m_c)
        0.00023950879616352339

        Now we consider the same situation again, this time for a two-
        compartment tissue with one central compartment that exchanges with 
        the enviroment. 

        >>> R1 = [0.5, 1.5]
        >>> v = [0.3, 0.6]
        >>> PS = 0.1
        
        Compute magnetization without inflow:

        >>> Fw = [[0, PS], [PS, 0]]
        >>> M_c = dc.Mz_ss(R1, TR, FA, v, Fw)

        Compute magnetization with flow through the first compartment:

        >>> Fw = [[f, PS], [PS, 0]]
        >>> M_f = np.zeros((2, 100))
        >>> for i, m in enumerate(mi):
        >>>     M_f[:,i] = dc.Mz_ss(R1, TR, FA, v, Fw, j=[f*m, 0])

        Plot the results for the central compartment:

        >>> plt.plot(mi, M_c[0]/v[0] + mi*0, label='No inflow', linewidth=3)
        >>> plt.plot(mi, M_f[0,:]/v[0], label='Inflow', linewidth=3)
        >>> plt.xlabel('Inflow magnetization (A/cm/mL)')
        >>> plt.ylabel('Steady-state magnetization (A/cm/mL)')
        >>> plt.legend()
        >>> plt.show()

        We can verify again that the magnetization is the same as that of the 
        isolated tissue when the inflow is at the isolated steady-state: 

        >>> m_c = M_c[0]/v[0]
        >>> M_f = dc.Mz_ss(R1, TR, FA, v, Fw=f, j=f*m_c)
        >>> print(M_f[0]/v[0] - m_c) 
        0.0002984954839493209      
     
    """
    # One compartment
    if np.isscalar(v):
        if np.isscalar(R1): 
            return me*_Nz_ss_1c(R1, TR, FA, v, Fw, j)
        else:
            R1 = np.array(R1)
            nt = np.size(R1)
            M = np.zeros(nt)
            for t in range(nt):
                jt = None if j is None else j[t]
                if np.isscalar(FA):
                    M[t] = Mz_ss(R1.reshape(-1)[t], TR, FA, v, Fw, jt, me)
                else:
                    M[t] = Mz_ss(R1.reshape(-1)[t], TR, FA.reshape(-1)[t], v, Fw, jt, me)
            return M.reshape(R1.shape)
    
    # Multiple compartments  
    else:
        
        n = len(v)
        if np.isscalar(R1):
            raise ValueError(
                'In a tissue with n compartments, R1 must have at least n elements.')
        R1 = np.array(R1)
        if R1.shape[0] != n:
            raise ValueError(
                'In a tissue with n compartments, the first dimension of R1 must have size n.')
        
        if np.ndim(R1)==1:
            return me*_Nz_ss(R1, TR, FA, v, Fw, j)
        
        shape = R1.shape
        R1 = R1.reshape(n, -1)
        if j is not None:
            j = j.reshape(n, -1)
        M = np.zeros(R1.shape)
        for t in range(R1.shape[1]):
            jt = None if j is None else j[:,t]
            M[:,t] = Mz_ss(R1[:,t], TR, FA, v, Fw, jt, me) 
        return M.reshape(shape)   


def _Nz_ss_1c(R1, TR, FA, v, Fw=0, j=None):
    K = R1 + Fw/v if v > 0 else R1
    J = R1*v if j is None else R1*v + j
    E = np.exp(-np.multiply(TR, K))
    cFA = np.cos(FA*np.pi/180)
    n = (1-E) / (1-cFA*E)
    if K==0:
        return n
    else:
        return n * J/K


def _Nz_ss(R1, TR, FA, v, Fw=0, j=None):   
    nc = np.size(R1)

    # Closed tissue with constant permeabilities
    if np.isscalar(Fw): 
        if Fw == np.inf: 
            return _Nz_ss_fex(R1, TR, FA, v)
        elif Fw == 0: 
            return _Nz_ss_nex(R1, TR, FA, v)
        else:
            Fw = np.full((nc, nc), Fw) - np.diag(np.full(nc, Fw))
            return _Nz_ss_aex(R1, TR, FA, v, Fw, j)
        
    # Open tissue
    if np.shape(Fw) != (nc, nc):
        raise ValueError('Fw must be a square array with size equal to the '
                         'number of compartments.')
    PSw = np.array(Fw) - np.diag(np.diag(Fw))
    ninf = np.count_nonzero(np.isinf(PSw))
    if np.linalg.norm(PSw) == 0:
        return _Nz_ss_nex(R1, TR, FA, v, Fw, j)
    elif ninf == nc*nc-nc:
        return _Nz_ss_fex(R1, TR, FA, v, Fw, j)
    elif 0 < ninf < nc*nc-nc:
        raise NotImplementedError(
            'Water exchange with some (but not all) infinite PS '
            'values is currently not implemented.')
    else:
        return _Nz_ss_aex(R1, TR, FA, v, Fw, j)


def _Nz_ss_fex(R1, TR, FA, v, Fw=0, j=None):
    R1 = np.sum(np.multiply(v, R1)) / np.sum(v)
    if j is None:
        N = _Nz_ss_1c(R1, TR, FA, np.sum(v))
    else:
        fo = np.diag(Fw)
        N = _Nz_ss_1c(R1, TR, FA, np.sum(v), np.sum(fo), np.sum(j))
    return np.stack([N*vc/np.sum(v) for vc in v])

def _Nz_ss_nex(R1, TR, FA, v, Fw=0, j=None):
    if j is None:
        N = [_Nz_ss_1c(R1[c], TR, FA, v[c]) for c in range(np.size(v))]
    else:
        N = []
        fo = np.diag(Fw)
        for c in range(np.size(v)):
            Nc = _Nz_ss_1c(R1[c], TR, FA, v[c], fo[c], j[c]) 
            N.append(Nc)
    return np.stack(N)
      
def _Nz_ss_aex(R1, TR, FA, v, Fw=0, j=None):
    K = _Mz_K(R1, v, Fw)
    J = _Mz_J(R1, v, 1, j=j)
    E = expm(-TR*K)
    I = np.eye(np.size(R1))
    cFA = np.cos(FA*np.pi/180)
    A = np.dot(K, I-cFA*E)
    Ainv = np.linalg.inv(A)
    return np.dot(Ainv, np.dot(I-E, J))        




def Mz_spgr(R1, T, TR, FA, TP=0, v=1, Fw=0, j=None, n0=0, me=1):
    """Longitudinal tissue magnetization of a spoiled gradient-echo sequence.

    See section :ref:`basics-relaxation-T1` for more detail.

    Args:
        R1 (array-like): Longitudinal relaxation rates in 1/sec. For a tissue 
          with n compartments, the first dimension of R1 must be n. For a
          single compartment, R1 can be scalar or a 1D time-array.
        T (float): time since the first rf-pulse.
        TR (float): repetition time between rf-pulses.
        FA (float): flip angle of the rf-pulse.
        TP (float, optional): preparation delay. Default is 0.
        v (array-like, optional): volume fractions of the compartments. For a 
          one-compartment tissue this is a scalar - otherwise it is an 
          array with one value for each compartment. Defaults to 1.
        Fw (array-like, optional): Water flow between the compartments and to 
          the environment, in units of mL/sec/cm3. Generally Fw must be a nxn 
          array, where n is the number of compartments, and the off-diagonal 
          elements Fw[j,i] are the permeability for water moving from 
          compartment i into j. The diagonal elements Fw[i,i] quantify the 
          flow of water from compartment i to outside. For a closed system 
          with equal permeabilities between all compartments, a scalar value 
          for Fw can be provided. Defaults to 0.
        j (array-like, optional): normalized tissue magnetization flux. j has 
          to have the same shape as R1. Defaults to None.
        n0 (array-like, optional): initial relative magnetization at T=0. 
          If this is a scalar, all compartments are assumed to have the same 
          initial magnetization. Defaults to 0.
        me (array-like, optional): equilibrium magnetization of the tissue 
          compartments. If a scalar value is provided, all compartments are 
          assumed to have the same equilibrium magnetization. Defaults to 1.

    Returns:
        np.ndarray: Tissue magnetization in the compartments after a time T.

    Example:

        Compare SPGR tissue magnetization of a two-compartment tissue
        after inversion against free recovery and steady-state:

    .. plot::
        :include-source:
        :context: close-figs

        Import packages:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Define constants:

        >>> FA, TR, TP = 12, 0.005, 0
        >>> TI = np.linspace(0,3,100)
        >>> R1 = [1, 0.5]
        >>> v = [0.3, 0.7]
        >>> f, PS = 0.5, 0.1
        >>> Fw = [[f, PS], [PS, 0]]

        Compute magnetization:

        >>> Mspgr = dc.Mz_spgr(R1, TI, TR, FA, TP, v, Fw, j=[f, 0], n0=-1) 
        >>> Mfree = dc.Mz_free(R1, TI, v, Fw, j=[f, 0], n0=-1)
        >>> Mss = dc.Mz_ss(R1, TR, FA, v, Fw, j=[f, 0])

        Plot the results for the peripheral compartment:

        >>> c = 1
        >>> plt.title('Peripheral compartment')
        >>> plt.plot(TI, Mfree[c,:], label='Free', linewidth=3)
        >>> plt.plot(TI, v[c]+0*TI, label='Equilibrium', linewidth=3)
        >>> plt.plot(TI, Mspgr[c,:], label='SPGR', linewidth=3)
        >>> plt.plot(TI, Mss[c]+0*TI, label='Steady-state', linewidth=3)
        >>> plt.xlabel('Time since start of pulse sequence (sec)')
        >>> plt.ylabel('Tissue magnetization (A/cm/cm3)')
        >>> plt.legend()
        >>> plt.show()

        This verifies that the free recovery magnetization goes to equilibrium, 
        and that the SPGR magnetization relaxes to the steady state at a 
        shorter time than the free recovery.

    """

    if not np.isscalar(T): # Do we need this?
        if np.isscalar(R1):
            M = np.empty(np.size(T))
            for k, Tk in enumerate(T):
                M[k] = Mz_spgr(R1, Tk, TR, FA, TP, v, Fw, j, n0, me)
        else:
            M = np.empty(np.shape(R1) + (np.size(T), ))
            for k, Tk in enumerate(T):
                M[...,k] = Mz_spgr(R1, Tk, TR, FA, TP, v, Fw, j, n0, me)
        return M

    # One compartment
    if np.isscalar(v):
        if np.isscalar(R1):
            nx = T/TR
            ncFA = np.cos(FA*np.pi/180)**nx
            M0 = n0*v*me
            Mss = Mz_ss(R1, TR, FA, v, Fw, j, me)
            K = _Mz_K(R1, v, Fw)
            E = np.exp(-T*K)
            return Mss + ncFA*E*(M0-Mss)
        else:
            R1 = np.array(R1)
            nt = np.size(R1)
            M = np.empty(nt)
            for t in range(nt):
                jt = None if j is None else j[t]
                n0t = n0 if np.isscalar(n0) else n0[t]
                M[t] = Mz_spgr(R1.reshape(-1)[t], T, TR, FA, TP, v, Fw, jt, n0t, me)
            return M

    # Multiple compartments  
    else:

        n = len(v)
        if np.isscalar(R1):
            raise ValueError(
                'In a tissue with n compartments, R1 must have at least n elements.')
        R1 = np.array(R1)
        if R1.shape[0] != n:
            raise ValueError(
                'In a tissue with n compartments, the first dimension of R1 must have size n.')

        if np.ndim(R1)==1:
            nx = T/TR
            ncFA = np.cos(FA*np.pi/180)**nx
            M0 = np.multiply(np.multiply(n0, v), me)
            Mss = Mz_ss(R1, TR, FA, v, Fw, j, me)
            K = _Mz_K(R1, v, Fw)
            E = expm(-T*K)
            return Mss + ncFA*np.dot(E, M0-Mss)

        shape = R1.shape
        R1 = R1.reshape(n, -1)
        if j is not None:
            j = j.reshape(n, -1)
        M = np.zeros(R1.shape)
        for t in range(R1.shape[1]):
            jt = None if j is None else j[:,t]
            if np.ndim(n0)==2:
                n0t = n0[:,t]
            else:
                n0t = n0
            M[:,t] = Mz_spgr(R1[:,t], T, TR, FA, TP, v, Fw, jt, n0t, me)
        return M.reshape(shape) 



def signal(sequence, R1, S0, TR=None, FA=None, TC=None):
    # Private for now - generalize to umbrella signal function
    if sequence == 'SRC':
        return signal_free(S0, R1, TC, FA)
    if sequence == 'SR':
        return signal_spgr(S0, R1, TC, TR, FA)
    elif sequence == 'SS':
        return signal_ss(S0, R1, TR, FA)
    else:
        raise ValueError(
            'Sequence ' + str(sequence) + ' is not recognised.')


def signal_dsc(R1, R2, S0: float, TR, TE) -> np.ndarray:
    """Signal model for a DSC scan with T2 and T2-weighting.

    Args:
        R1 (array-like): Longitudinal relaxation rate in 1/sec. Must have the same size as R2.
        R2 (array-like): Transverse relaxation rate in 1/sec. Must have the same size as R1.
        S0 (float): Signal scaling factor (arbitrary units).
        TR (array-like): Repetition time, or time between successive selective excitations, in sec. If TR is an array, it must have the same size as R1 and R2.
        TE (array-like): Echo time, in sec. If TE is an array, it must have the same size as R1 and R2.

    Returns:
        np.ndarray: Signal in arbitrary units, same length as R1 and R2.
    """
    return S0*np.exp(-np.multiply(TE, R2))*(1-np.exp(-np.multiply(TR, R1)))


def signal_t2w(R2, S0: float, TE) -> np.ndarray:
    """Signal model for a DSC scan with T2-weighting.

    Args:
        R2 (array-like): Transverse relaxation rate in 1/sec. Must have the same size as R1.
        S0 (float): Signal scaling factor (arbitrary units).
        TE (array-like): Echo time, in sec. If TE is an array, it must have the same size as R1 and R2.

    Returns:
        np.ndarray: Signal in arbitrary units, same length as R1 and R2.
    """
    return S0*np.exp(-np.multiply(TE, R2))


def signal_free(S0, R1, T, FA, v=1, Fw=0, j=None, n0=0, R10=None, noise_sdev=0):
    """Signal with readout after a free recovery.

    Args:
        S0 (float): Signal scaling factor (arbitrary units).
        R1 (array-like): Longitudinal relaxation rates in 1/sec. For a tissue 
          with n compartments, the first dimension of R1 must be n. For a 
          tissue with a single compartment, R1 can have any shape. If R1 is a 
          scalar, the result for a well-mixed tissue is returned.
        T (float): time to center of k-space
        FA (float): flip angle
        v (array-like, optional): volume fractions of the compartments. For a 
          one-compartment tissue this is a scalar - otherwise it is an 
          array with one value for each compartment. Defaults to 1.
        Fw (array-like, optional): Water flow between the compartments and to 
          the environment, in units of mL/sec/cm3. Generally Fw must be a nxn 
          array, where n is the number of compartments, and the off-diagonal 
          elements Fw[j,i] are the permeability for water moving from 
          compartment i into j. The diagonal elements Fw[i,i] quantify the 
          flow of water from compartment i to outside. For a closed system 
          with equal permeabilities between all compartments, a scalar value 
          for Fw can be provided. Defaults to 0.
        j (array-like, optional): normalized tissue magnetization flux. j has 
          to have the same shape as R1. Defaults to None.
        n0 (array-like, optional): initial relative magnetization at T=0. 
          If this is a scalar, all compartments are assumed to have the same 
          initial magnetization. Defaults to 0.
        R10 (float, optional): R1-value where S0 is defined. If not provided, 
          S0 is the scaling factor corresponding to infinite R10. Defaults 
          to None.
        noise_sdev (float, optional): standard deviation of the signal noise. 
          If this is non-zero, the signal is computed from the mean of the 
          Rician distributions. A noise_sdev=0 has no effect.

    Returns:
        np.ndarray: Signal in the same units as S0 and with the same 
        dimensions as R1.
    """
    if R10 is not None:
        S0 = S0/signal_free(1, R10, T, FA, v, Fw, j, n0)
    Mz = Mz_free(R1, T, v, Fw, j, n0)
    sFA = np.sin(FA*np.pi/180)

    # One compartment
    if np.isscalar(v):
        return _signal_rice(S0 * sFA * np.abs(Mz), noise_sdev)
    
    # Multiple compartments - sum over compartments (first dim)
    else:
        return _signal_rice(S0 * sFA * np.abs(np.sum(Mz, axis=0)), noise_sdev)


def signal_ss(S0, R1, TR, FA, v=1, Fw=0, j=None, R10=None, noise_sdev=0) -> np.ndarray:
    """Signal of a spoiled gradient echo sequence applied in steady state.

    Args:
        S0 (float): Signal scaling factor (arbitrary units).
        R1 (array-like): Longitudinal relaxation rates in 1/sec. For a tissue 
          with n compartments, the first dimension of R1 must be n. For a 
          tissue with a single compartment, R1 can have any shape. If R1 is a 
          scalar, the result for a well-mixed tissue is returned.
        TR (float): Repetition time, or time between successive selective 
          excitations, in sec. 
        FA (float): Flip angle in degrees.
        v (array-like, optional): volume fractions of the compartments. For a 
          one-compartment tissue this is a scalar - otherwise it is an 
          array with one value for each compartment. Defaults to 1.
        Fw (array-like, optional): Water flow between the compartments and to 
          the environment, in units of mL/sec/cm3. Generally Fw must be a nxn 
          array, where n is the number of compartments, and the off-diagonal 
          elements Fw[j,i] are the permeability for water moving from 
          compartment i into j. The diagonal elements Fw[i,i] quantify the 
          flow of water from compartment i to outside. For a closed system 
          with equal permeabilities between all compartments, a scalar value 
          for Fw can be provided. Defaults to 0.
        j (array-like, optional): normalized tissue magnetization flux. j has 
          to have the same shape as R1. Defaults to None.
        R10 (float, optional): R1-value where S0 is defined. If not provided, 
          S0 is the scaling factor corresponding to infinite R10. Defaults 
          to None.
        noise_sdev (float, optional): standard deviation of the signal noise. 
          If this is non-zero, the signal is computed from the mean of the 
          Rician distributions. A noise_sdev=0 has no effect.

    Returns:
        np.ndarray: Signal in the same units as S0
    """
    if R10 is not None:
        S0 = S0/signal_ss(1, R10, TR, FA, v, Fw, j)
    Mz = Mz_ss(R1, TR, FA, v, Fw, j)
    
    sFA = np.sin(FA*np.pi/180)
    # One compartment
    if np.isscalar(v):
        return _signal_rice(S0 * sFA * np.abs(Mz), noise_sdev)
    
    # Multiple compartments - sum over compartments (first dim)
    else:
        return _signal_rice(S0 * sFA * np.abs(np.sum(Mz, axis=0)), noise_sdev)


def signal_spgr(S0, R1, T, TR, FA, TP=0.0, 
                  v=1, Fw=0, j=None, n0=0, R10=None, noise_sdev=0) -> np.ndarray:
    """Signal of an SPGR sequence.

    Args:
        S0 (float): Signal scaling factor (arbitrary units).
        R1 (array-like): Longitudinal relaxation rate in 1/sec.
        T (float): time since the first readout rf-pulse.
        TR (float): Repetition time, or time between successive selective 
          excitations, in sec. 
        FA (array-like): Flip angle in degrees.
        TP (float, optional): Time (sec) between the preparation pre-pulse and 
          the first readout pulse. Defaults to 0.
        v (array-like, optional): volume fractions of the compartments. For a 
          one-compartment tissue this is a scalar - otherwise it is an 
          array with one value for each compartment. Defaults to 1.
        Fw (array-like, optional): Water flow between the compartments and to 
          the environment, in units of mL/sec/cm3. Generally Fw must be a nxn 
          array, where n is the number of compartments, and the off-diagonal 
          elements Fw[j,i] are the permeability for water moving from 
          compartment i into j. The diagonal elements Fw[i,i] quantify the 
          flow of water from compartment i to outside. For a closed system 
          with equal permeabilities between all compartments, a scalar value 
          for Fw can be provided. Defaults to 0.
        j (array-like, optional): normalized tissue magnetization flux. j has 
          to have the same shape as R1. Defaults to None.
        n0 (array-like, optional): initial relative magnetization at T=0. 
          If this is a scalar, all compartments are assumed to have the same 
          initial magnetization. Defaults to 0.
        R10 (float, optional): R1-value where S0 is defined. If not provided, 
          S0 is the scaling factor corresponding to infinite R10. Defaults to 
          None.
        noise_sdev (float, optional): standard deviation of the signal noise. 
          If this is non-zero, the signal is computed from the mean of the 
          Rician distributions. A noise_sdev=0 has no effect.

    Raises:
        ValueError: If TP is larger than TC.

    Returns:
        np.ndarray: Signal in arbitrary units, of the same length as R1.
    """
    if R10 is not None:
        S0 = S0/signal_spgr(1, R10, T, TR, FA, TP, v, Fw, j, n0)
    if TP>0:
        N0 = Mz_free(R1, TP, v, Fw, j, n0)
        if np.ndim(N0)==2:
            n0 = N0
            for t in range(N0.shape[1]):
                n0[:,t] = np.divide(N0[:,t], v)
        else:
            n0 = np.divide(N0, v)
    Mz = Mz_spgr(R1, T, TR, FA, TP, v, Fw, j, n0)

    sFA = np.sin(FA*np.pi/180)
    # One compartment
    if np.isscalar(v):
        return _signal_rice(S0 * sFA * np.abs(Mz), noise_sdev)
    
    # Multiple compartments - sum over compartments (first dim)
    else:
        return _signal_rice(S0 * sFA * np.abs(np.sum(Mz, axis=0)), noise_sdev)


def signal_src(S0, R1, TC, noise_sdev=0)-> np.ndarray:
    """Signal of a saturation-recovery sequence with a center-encoded readout.

    Args:
        S0 (float): Signal scaling factor (arbitrary units).
        R1 (array-like): Longitudinal relaxation rate in 1/sec.
        TC (float): Time (sec) between the saturation pulse and the acquisition of the k-space center.
        noise_sdev (float, optional): standard deviation of the signal noise. 
          If this is non-zero, the signal is computed from the mean of the 
          Rician distributions. A noise_sdev=0 has no effect.

    Returns:
        np.ndarray: Signal in arbitrary units, same length as R1.
    """
    return _signal_rice(S0 * (1 - np.exp(-TC * R1)), noise_sdev)


def signal_lin(S0, R1, noise_sdev=0) -> np.ndarray:
    """Signal for any sequence operating in the linear regime.

    Args:
        S0 (float): Signal scaling factor (arbitrary units).
        R1 (array-like): Longitudinal relaxation rate in 1/sec.
        noise_sdev (float, optional): standard deviation of the signal noise. 
          If this is non-zero, the signal is computed from the mean of the 
          Rician distributions. A noise_sdev=0 has no effect.
        
    Returns:
        np.ndarray: Signal in arbitrary units, of the same length as R1.
    """
    return _signal_rice(S0 * R1, noise_sdev)





def conc_t2w(S, TE: float, r2=0.5, n0=1) -> np.ndarray:
    """Concentration for a DSC scan with T2-weighting.

    Args:
        S (array-like): Signal in arbitrary units.
        TE (float): Echo time in sec.
        r2 (float, optional): Transverse relaxivity in Hz/M. Defaults to 0.5.
        n0 (int, optional): Baseline length. Defaults to 1.

    Returns:
        np.ndarray: Concentration in M, same length as S.
    """
    # S/Sb = exp(-TE(R2-R2b))
    #   ln(S/Sb) = -TE(R2-R2b)
    #   R2-R2b = -ln(S/Sb)/TE
    # R2 = R2b + r2C
    #   C = (R2-R2b)/r2
    #   C = -ln(S/Sb)/TE/r2
    Sb = np.mean(S[:n0])
    C = -np.log(S/Sb)/TE/r2
    return C



def conc_ss(S, TR: float, FA: float, T10: float, r1=0.005, n0=1, S0=None) -> np.ndarray:
    """Concentration of a spoiled gradient echo sequence applied in steady state.

    Args:
        S (array-like): Signal in arbitrary units.
        TR (float): Repetition time, or time between successive selective excitations, in sec.
        FA (float): Flip angle in degrees.
        T10 (array-like): baseline T1 value in sec.
        r1 (float, optional): Longitudinal relaxivity in Hz/M. Defaults to 0.005.
        n0 (int, optional): Baseline length. Defaults to 1.

    Returns:
        np.ndarray: Concentration inM , same length as S.
    """
    # S = Sinf * (1-exp(-TR*R1)) / (1-cFA*exp(-TR*R1))
    # Sb = Sinf * (1-exp(-TR*R10)) / (1-cFA*exp(-TR*R10))
    # Sn = (1-exp(-TR*R1)) / (1-cFA*exp(-TR*R1))
    # Sn * (1-cFA*exp(-TR*R1)) = 1-exp(-TR*R1)
    # exp(-TR*R1) - Sn *cFA*exp(-TR*R1) = 1-Sn
    # (1-Sn*cFA) * exp(-TR*R1) = 1-Sn
    S = np.array(S)
    shape = S.shape
    S = S.reshape(-1, shape[-1])
    if np.isscalar(T10):
        T10 = np.full(S.shape[0], T10)
    else:
        T10 = T10.reshape(S.shape[0])
    with np.errstate(divide="ignore", invalid="ignore", over="ignore"):
        R10 = np.where(T10==0, 0, 1/T10)

    cFA = np.cos(FA*np.pi/180)
    if S0 is None:
        # If S0 is not provided, estimate from the data
        Sb = np.mean(S[:, :n0], axis=-1)
        Sb = np.broadcast_to(Sb[..., None], S.shape)
        R10 = np.broadcast_to(R10[..., None], S.shape)
        Sn = np.zeros(S.shape)
        nozero = Sb > 0
        E0 = np.exp(-TR*R10)
        E0 = E0[nozero]
        Sn[nozero] = (S[nozero]/Sb[nozero])*(1-E0)/(1-cFA*E0)	     
        # Replace any Nan values by interpolating between nearest neighbours
        outrange = Sn >= 1
        if np.sum(outrange) > 0:
            inrange = Sn < 1
            x = np.arange(Sn.size).reshape(Sn.shape)
            Sn[outrange] = np.interp(x[outrange], x[inrange], Sn[inrange])
    else:
        S0 = np.array(S0).reshape(-1)
        S0 = np.broadcast_to(S0[..., None], S.shape)
        with np.errstate(divide="ignore", invalid="ignore", over="ignore"):
            Sn = np.where(S0>0, S/S0/np.sin(np.deg2rad(FA)), 0)
        R10 = np.broadcast_to(R10[..., None], S.shape)
    Sn = (1-Sn)/(1-cFA*Sn)
    with np.errstate(divide="ignore", invalid="ignore", over="ignore"):
        R1 = np.where(Sn==0, 0, -np.log(Sn)/TR)  
    C = (R1 - R10)/r1
    return C.reshape(shape)


def conc_src(S, TC: float, T10: float, r1=0.005, n0=1, S0=None) -> np.ndarray:
    """Concentration of a saturation-recovery sequence with a center-encoded readout.

    Args:
        S (array-like): Signal in arbitrary units.
        TC (float): Time (sec) between the saturation pulse and the acquisition of the k-space center.
        T10 (float): baseline T1 value in sec.
        r1 (float, optional): Longitudinal relaxivity in Hz/M. Defaults to 0.005.
        n0 (int, optional): Baseline length. Defaults to 1.

    Returns:
        np.ndarray: Concentration in M, same length as S.

    Example:

        We generate some signals from ground-truth concentrations, then reconstruct the concentrations and check against the ground truth:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        First define some constants:

        >>> T10 = 1         # sec
        >>> TC = 0.2        # sec
        >>> r1 = 0.005      # Hz/M
        >>> FA = 15         # deg

        Generate ground truth concentrations and signal data:

        >>> t = np.arange(0, 5*60, 0.1)     # sec
        >>> C = 0.003*(1-np.exp(-t/60))     # M
        >>> R1 = 1/T10 + r1*C               # Hz
        >>> S = dc.signal_free(100, R1, TC, FA)  # au

        Reconstruct the concentrations from the signal data:

        >>> Crec = dc.conc_src(S, TC, T10, r1)

        Check results by plotting ground truth against reconstruction:

        >>> plt.plot(t/60, 1000*C, 'ro', label='Ground truth')
        >>> plt.plot(t/60, 1000*Crec, 'b-', label='Reconstructed')
        >>> plt.title('SRC signal inverse')
        >>> plt.xlabel('Time (min)')
        >>> plt.ylabel('Concentration (mM)')
        >>> plt.legend()
        >>> plt.show()

    """
    # S = S0*(1-exp(-TC*R1))
    # S/Sb = (1-exp(-TC*R1))/(1-exp(-TC*R10))
    # (1-exp(-TC*R10))*S/Sb = 1-exp(-TC*R1)
    # 1-(1-exp(-TC*R10))*S/Sb = exp(-TC*R1)
    # ln(1-(1-exp(-TC*R10))*S/Sb) = -TC*R1
    # -ln(1-(1-exp(-TC*R10))*S/Sb)/TC = R1
    S = np.array(S)
    shape = S.shape
    S = S.reshape(-1, shape[-1])
    if np.isscalar(T10):
        T10 = np.full(S.shape[0], T10)
    else:
        T10 = T10.reshape(S.shape[0])
    with np.errstate(divide="ignore", invalid="ignore", over="ignore"):
        R10 = np.where(T10==0, 0, 1/T10)
    R1 = np.zeros(S.shape)
    if S0 is None:
        Sb = np.mean(S[:, :n0], axis=-1)
        Sb = np.broadcast_to(Sb[..., None], S.shape)
        R10 = np.broadcast_to(R10[..., None], S.shape)
        E = np.exp(-TC*R10)
        nozero = Sb > 0
        R1[nozero] = -np.log(1-(1-E[nozero])*S[nozero]/Sb[nozero])/TC
    else:
        nozero = S0 > 0
        R1[nozero] = - np.log(1 - S[nozero] / S0[nozero]) / TC
        R10 = np.broadcast_to(R10[..., None], S.shape)
    C = (R1 - R10)/r1
    return C.reshape(shape)


def conc_lin(S, T10, r1=0.005, n0=1, S0=None):
    """Concentration for any sequence operating in the linear regime.

    Args:
        S (array-like): Signal in arbitrary units.
        T10 (float): baseline T1 value in sec.
        r1 (float, optional): Longitudinal relaxivity in Hz/M. Defaults to 0.005.
        n0 (int, optional): Baseline length. Defaults to 1.

    Returns:
        np.ndarray: Concentration in M, same length as S.
    """
    # S = S0 * R10
    S = np.array(S)
    shape = S.shape
    S = S.reshape(-1, shape[-1])
    if np.isscalar(T10):
        T10 = np.full(S.shape[0], T10)
    else:
        T10 = T10.reshape(S.shape[0])
    T10 = np.broadcast_to(T10[..., None], S.shape)
    with np.errstate(divide="ignore", invalid="ignore", over="ignore"):
        R10 = np.where(T10==0, 0, 1/T10)
    R1 = np.zeros(S.shape)
    if S0 is None:
        Sb = np.mean(S[:, :n0], axis=-1)
        Sb = np.broadcast_to(Sb[..., None], S.shape)
        nozero = Sb > 0
        R1[nozero] = R10[nozero]*S[nozero]/Sb[nozero]  # relaxation rate in 1/msec
    else:
        S0 = S0.reshape(S.shape[0])
        S0 = np.broadcast_to(S0[..., None], S.shape)
        nozero = S0 > 0
        R1[nozero] = S[nozero] / S0[nozero]
    C = (R1 - R10)/r1
    return C.reshape(shape)


